import json
import logging
from pathlib import Path

import h5py
import numpy as np
import torch
from torch.nn import CrossEntropyLoss
from torch.optim.adam import Adam
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import ListOfBlobs, Session, __version__, conf
from idtrackerai.utils import create_dir, load_id_images

from ..network import (
    DEVICE,
    IdCNN,
    IdentifierIdCNN,
    StopTraining,
    get_dataloader,
    get_predictions,
    train_loop,
)
from .crossings_dataset import get_train_validation_and_eval_blobs
from .model_area import ModelArea


def _apply_area_and_unicity_heuristics(
    list_of_blobs: ListOfBlobs, n_animals: int
) -> None:
    logging.info(
        "Classifying Blobs as individuals or crossings "
        "depending on their area and the number of blobs in the frame"
    )

    model_area = ModelArea(list_of_blobs, n_animals)

    for blobs_in_frame in list_of_blobs.blobs_in_video:
        unicity_cond = len(blobs_in_frame) == n_animals
        for blob in blobs_in_frame:
            blob.seems_like_individual = unicity_cond or model_area(blob.area)

    n_seems_like_individual = sum(
        blob.seems_like_individual for blob in list_of_blobs.all_blobs
    )
    logging.info(
        f"{n_seems_like_individual} blobs seem like individuals, "
        f"{list_of_blobs.number_of_blobs - n_seems_like_individual} seem like crossings"
    )


def _update_id_image_dataset_with_crossings(
    list_of_blobs: ListOfBlobs, id_images_file_paths: list[Path]
):
    """Adds a array to the identification images files indicating whether
    each image is an individual or a crossing.
    """
    logging.info("Updating crossings in identification images files")

    crossings = []
    for path in id_images_file_paths:
        with h5py.File(path, "r") as file:
            crossings.append(np.empty(len(file["id_images"]), bool))  # type: ignore

    for blob in list_of_blobs.all_blobs:
        crossings[blob.episode][blob.id_image_index] = blob.is_a_crossing

    for path, crossing in zip(id_images_file_paths, crossings):
        try:
            with h5py.File(path, "r+") as file:
                if "crossings" in file:  # in case we are reusing the file in API calls
                    logging.warning("Overwriting existing crossings in %s", path)
                    del file["crossings"]
                file.create_dataset("crossings", data=crossing)
        except BlockingIOError as exc:
            # Some MacOS crash with
            # BlockingIOError: [Errno 35] Unable to open file (unable to lock file, errno = 35, error message = 'Resource temporarily unavailable')
            logging.error(f"Failed at writing in {path}: {exc}")


def detect_crossings(list_of_blobs: ListOfBlobs, session: Session) -> None:
    """Classify all blobs in the video as being crossings or individuals"""

    _apply_area_and_unicity_heuristics(list_of_blobs, session.n_animals)

    train_images, train_labels, train_weights, val_images, val_labels = (
        get_train_validation_and_eval_blobs(
            list_of_blobs.blobs_in_video, session.n_animals
        )
    )

    unknown_blobs = [
        blob
        for blob in list_of_blobs.all_blobs
        if not hasattr(blob, "is_an_individual")
    ]

    logging.info(f"{len(unknown_blobs)} unknown blobs")

    if (
        np.bincount(train_labels, minlength=2).min()
        < conf.MINIMUM_NUMBER_OF_CROSSINGS_TO_TRAIN_CROSSING_DETECTOR
    ):
        logging.debug("There are not enough crossings to train the crossing detector")
        for blob in unknown_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return
    logging.info("There are enough crossings to train the crossing detector")

    create_dir(session.crossings_detector_folder, remove_existing=True)

    train_loader = get_dataloader(
        "training",
        load_id_images(session.id_images_file_paths, train_images),
        train_labels,
        conf.BATCH_SIZE_DCD,
    )

    val_loader = get_dataloader(
        "validation",
        load_id_images(session.id_images_file_paths, val_images),
        val_labels,
    )

    crossing_model = IdCNN(input_shape=session.id_image_size, out_dim=2).to(DEVICE)
    optimizer = Adam(crossing_model.parameters(), lr=conf.LEARNING_RATE_DCD)
    scheduler = MultiStepLR(optimizer, milestones=[30, 60], gamma=0.1)
    criterion = CrossEntropyLoss(torch.tensor(train_weights, dtype=torch.float32)).to(
        DEVICE
    )
    stopping = StopTraining(
        epochs_limit=conf.MAXIMUM_NUMBER_OF_EPOCHS_DCD,
        overfitting_limit=conf.OVERFITTING_COUNTER_THRESHOLD_DCD,
        plateau_limit=conf.LEARNING_RATIO_DIFFERENCE_DCD,
    )
    if conf.TORCH_COMPILE:
        crossing_model.compile()
        criterion.compile()

    with (session.crossings_detector_folder / "model_params.json").open("w") as file:
        json.dump(
            {
                "n_classes": 2,
                "image_size": session.id_image_size,
                "resolution_reduction": session.resolution_reduction,
                "model": crossing_model.__class__.__name__,
                "version": __version__,
            },
            file,
            indent=4,
        )

    try:
        train_loop(
            crossing_model,
            criterion,
            optimizer,
            train_loader,
            val_loader,
            stopping,
            scheduler,
        )
    except RuntimeError as exc:
        logging.warning(
            "[red]The model diverged[/] provably due to a bad segmentation. Falling"
            " back to individual-crossing discrimination by average area model."
            " Original error: %s",
            exc,
            extra={"markup": True},
        )
        for blob in unknown_blobs:
            blob.is_an_individual = blob.seems_like_individual
        return

    del train_loader
    del val_loader

    model_path = session.crossings_detector_folder / "crossing_detector.model.pt"
    logging.info("Saving model at %s", model_path)
    torch.save(crossing_model.state_dict(), model_path)

    logging.info("Using crossing detector to classify individuals and crossings")
    predictions, _softmax = get_predictions(
        IdentifierIdCNN(crossing_model),
        [(blob.id_image_index, blob.episode) for blob in unknown_blobs],
        session.id_images_file_paths,
        "crossings",
    )

    logging.info(
        "Prediction results: %d individuals and %d crossings",
        np.count_nonzero(predictions == 1),
        np.count_nonzero(predictions == 2),
    )
    for blob, prediction in zip(unknown_blobs, predictions):
        blob.is_an_individual = prediction != 2

    _update_id_image_dataset_with_crossings(list_of_blobs, session.id_images_file_paths)



"Manage trajectory files saving and loading. In all compatible formats  and versions"

import json
import logging
import pickle
from argparse import ArgumentParser
from collections.abc import Iterable
from contextlib import suppress
from pathlib import Path
from typing import Literal

import numpy as np
from h5py import Dataset, File, Group

from . import create_dir, json_default, resolve_path, wrap_entrypoint


def save_trajectories(
    path: Path,
    data: dict,
    formats: Iterable[Literal["h5", "hdf5", "npy", "numpy", "csv", "json", "pickle"]],
) -> None:
    """Save trajectory dict into files following the format specifications"""
    for format in formats:
        format = format.lower()
        if format in ("npy", "numpy"):
            _save_trajectories_into_np(path, data)
        elif format in ("h5", "hdf5"):
            _save_trajectories_into_h5(path, data)
        elif format in ("csv", "json"):
            _save_trajectories_into_csv(path, data)
        elif format == "pickle":
            _save_trajectories_into_pickle(path, data)
        else:
            logging.error(f"Not recognized trajectory format {format!r}")


def _save_trajectories_into_np(path: Path, data: dict) -> None:
    path = path / "trajectories.npy"
    logging.info(f"Saving trajectories in {path}")
    np.save(path, data)  # type: ignore


def _save_trajectories_into_pickle(path: Path, data: dict) -> None:
    path = path / "trajectories.pickle"
    logging.info(f"Saving trajectories in {path}")
    with path.open("wb") as file:
        pickle.dump(data, file)


def _save_trajectories_into_csv(path: Path, data: dict) -> None:
    path = path / "trajectories_csv"
    logging.info(f"Saving trajectories in {path}")
    create_dir(path, remove_existing=True)

    attributes_dict = {}
    for key, value in data.items():
        if key in ("trajectories", "id_probabilities"):
            _save_array_to_csv(
                path / (key + ".csv"),
                value,
                key=key,
                fps=data.get("frames_per_second", 1),
            )
        elif key == "areas":
            np.savetxt(
                path / (key + ".csv"),
                np.asarray((value["mean"], value["median"], value["std"])).T,
                delimiter=", ",
                header="mean, median, standard_deviation",
                fmt="%.1f",
                comments="",
            )
        else:
            attributes_dict[key] = value

    json.dump(
        attributes_dict,
        (path / "attributes.json").open("w"),
        indent=4,
        default=json_default,
    )


def _save_trajectories_into_h5(path: Path, data: dict) -> None:
    path = path / "trajectories.h5"
    logging.info(f"Saving trajectories in {path}")
    data_ = data.copy()
    with File(path, "w") as file:
        areas: dict = data_.pop("areas")
        areas_group = file.create_group("areas")
        for key, value in areas.items():
            areas_group.create_dataset(key, data=value)

        setup_points: dict = data_.pop("setup_points")
        setup_points_group = file.create_group("setup_points")
        for key, value in setup_points.items():
            setup_points_group.create_dataset(key, data=value)

        identities_groups: dict = data_.pop("identities_groups")
        identities_groups_group = file.create_group("identities_groups")
        for key, value in identities_groups.items():
            identities_groups_group.create_dataset(key, data=np.asarray(value))

        # length_unit can be None
        file.attrs["length_unit"] = data_.pop("length_unit") or -1

        for key, value in data_.items():
            if value is None:
                continue
            try:
                if isinstance(value, (int, float, str)) or (
                    isinstance(value, list) and isinstance(value[0], str)
                ):
                    file.attrs[key] = value
                else:
                    compression = "gzip" if isinstance(value, np.ndarray) else None
                    file.create_dataset(key, data=value, compression=compression)
            except Exception as exc:
                logging.error(
                    f'Error saving "{key}" of type {type(value).__name__}. Error: {exc}'
                )


def _save_array_to_csv(
    path: Path, array: np.ndarray, key: str, fps: float | None
) -> None:
    array = array.squeeze()
    if key == "id_probabilities":
        fmt = "%.3e"
    elif key == "trajectories":
        fmt = "%.3f"
    else:
        fmt = "%.3f"

    if array.ndim == 3:
        array_header = ",".join(
            coord + str(i) for i in range(1, array.shape[1] + 1) for coord in ("x", "y")
        )
        array = array.reshape((-1, array.shape[1] * array.shape[2]))
    elif array.ndim == 2:
        array_header = ",".join(f"{key}{i}" for i in range(1, array.shape[1] + 1))
    else:
        array_header = f"{key}1"

    fmt = [fmt] * (array.shape[1] if array.ndim > 1 else 1)

    if fps is not None:  # add time column
        array_header = "time," + array_header
        fmt = ["%.3f"] + fmt
        time = np.arange(len(array), dtype=float) / fps
        array = np.column_stack((time, array))

    np.savetxt(path, array, delimiter=",", header=array_header, fmt=fmt, comments="")


def load_trajectories(path: Path | str) -> dict:
    """Loads trajectories from the specific path.

    Parameters
    ----------
    path : Path | str
        Path to trajectories. This can be a session folder, a trajectories
        folder or a specific file in any of the accepted file formats

    Returns
    -------
    dict
        Trajectories dict

    Raises
    ------
    FileNotFoundError
        If `path` does not exist
    """
    path = resolve_path(path)
    if not path.exists():
        raise FileNotFoundError(f"{path} not found")

    if path.is_file() or (path.is_dir() and path.name.endswith("_csv")):
        return _load_trajectories_file(path)

    if path.is_dir() and path.name.startswith("session_"):
        path = path / "trajectories"

    for name in (
        "trajectories",
        "validated",
        "without_gaps",
        "with_gaps",
        "trajectories_validated",
        "trajectories_wo_gaps",
        "trajectories_wo_identification",
    ):
        for format in (".h5", ".hdf5", ".npy", ".pickle", "_csv"):
            with suppress(FileNotFoundError):
                return _load_trajectories_file(path / (name + format))

    raise FileNotFoundError(f"Could not find trajectories in {path}")


def _load_trajectories_file(path: Path) -> dict:
    if path.suffix in (".h5", ".hdf5"):
        out = _load_trajectories_from_h5(path)
    elif path.suffix == ".pickle":
        out = _load_trajectories_from_pickle(path)
    elif path.suffix in (".npy", ".numpy"):
        out = _load_trajectories_from_np(path)
    elif path.is_dir() and path.name.endswith("_csv"):
        out = _load_trajectories_from_csv(path)
    else:
        raise FileNotFoundError(f'Could not load trajectory file "{path}"')
    logging.info(f"Trajectories loaded from {path}")
    return out


def _load_trajectories_from_np(path: Path) -> dict:
    return np.load(path, allow_pickle=True).item()


def _load_trajectories_from_pickle(path: Path) -> dict:
    with open(path, "rb") as file:
        return pickle.load(file)


def _load_trajectories_from_h5(path: Path) -> dict:
    def recursively_load_h5_group(group: Group) -> dict:
        ans = {}
        ans.update(group.attrs)

        for key, item in group.items():
            if isinstance(item, Dataset):
                ans[key] = item[()]
            elif isinstance(item, Group):
                ans[key] = recursively_load_h5_group(item)
            else:
                logging.error(f"Not recognized H5DF object '{key}': {item}")
        return ans

    with File(path, "r") as file:
        return recursively_load_h5_group(file)


def _load_trajectories_from_csv(path: Path) -> dict:
    out = json.loads((path / "attributes.json").read_text())

    for array_file in path.glob("*.csv"):
        # read csv without the first column containing time
        out[array_file.stem] = np.loadtxt(array_file, skiprows=1, delimiter=",")[:, 1:]

    if "trajectories" in out:
        # reshape trajectories
        out["trajectories"] = out["trajectories"].reshape(
            len(out["trajectories"]), -1, 2
        )

    return out


@wrap_entrypoint
def idtrackerai_format() -> None:
    """Script to convert trajectory formats"""
    parser = ArgumentParser()

    parser.add_argument(
        "paths",
        help=(
            "Paths to trajectory files to convert. Can be session folders "
            "(to convert all files inside trajectory subfolder), arbitrary folder "
            "(to convert all files in it) and specific files."
        ),
        type=Path,
        nargs="+",
    )
    parser.add_argument(
        "--formats",
        help="A sequence of strings defining in which formats the trajectories should be saved",
        type=str,
        choices=["h5", "npy", "csv", "pickle"],
        nargs="+",
        required=True,
    )

    args = parser.parse_args()

    for path in args.paths:
        path = resolve_path(path)
        if not path.exists():
            logging.error(f"{path} does not exist")
            continue

        trajectories = load_trajectories(path)

        if path.is_file():
            save_path = path.parent
        elif path.name.startswith("session_"):
            save_path = path / "trajectories"
        else:
            save_path = path

        save_trajectories(save_path, trajectories, args.formats)


if __name__ == "__main__":
    idtrackerai_format()



import ast
from argparse import ArgumentParser, ArgumentTypeError, _ArgumentGroup
from collections.abc import Callable
from pathlib import Path

from idtrackerai import Session, __version__, conf
from idtrackerai.utils import resolve_path


def Bool(value: str) -> bool:
    valid = {"true": True, "t": True, "1": True, "false": False, "f": False, "0": False}

    lower_value = value.lower()
    if lower_value not in valid:
        raise ValueError
    return valid[lower_value]


def path(value: str) -> Path:
    return_path = resolve_path(value)
    if not return_path.exists():
        raise ArgumentTypeError(f"No such file or directory: {return_path}")
    return return_path


def pair_of_ints(value: str):
    out = ast.literal_eval(value)
    if not isinstance(out, (tuple, list)):
        raise ValueError

    out = list(out)
    if len(out) != 2:
        raise ValueError
    if any(not isinstance(x, int) for x in out):
        raise ValueError
    return out


def get_parser(defaults: dict | None = None) -> ArgumentParser:
    defaults = defaults or {}

    parser = ArgumentParser(
        prog="idtracker.ai",
        epilog="For more info visit https://idtracker.ai",
        exit_on_error=False,
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parsers: dict[str, ArgumentParser | _ArgumentGroup] = {"base": parser}

    def add_argument(
        name: str, help: str, type: Callable, group: str = "General", **kwargs
    ) -> None:
        name = name.lower()

        metavar = f"<{type.__name__.lower()}>"

        if "choices" in kwargs:
            help += f' (choices: {", ".join(map(str, kwargs["choices"]))})'

        if name in ("load", "session") or "(default: " in help:
            # Video has a load method, it's not the default for --load
            # name has an adaptative default value
            pass
        elif name.upper() in defaults:
            help += f" (default: {defaults[name.upper()]})"
        elif name.lower() in defaults:
            help += f" (default: {defaults[name.lower()]})"

        if group not in parsers:
            parsers[group] = parsers["base"].add_argument_group(group)

        if not help.endswith("."):
            help += "."
        parsers[group].add_argument(
            "--" + name, help=help, type=type, metavar=metavar, **kwargs
        )

    # General
    add_argument(
        "load",
        help=(
            "A list of .toml files to load session parameters in increasing priority order. "
            "Later files override earlier ones for the same parameter."
        ),
        type=path,
        nargs="*",
        dest="parameters",
    )
    parsers["General"].add_argument(
        "--track",
        help="Start tracking directly from the terminal, bypassing the graphical user interface (GUI).",
        action="store_true",
    )
    add_argument(
        "name",
        help="Name of the session. Defaults to the name of the video files if not specified.",
        type=str,
    )
    add_argument(
        "video_paths",
        help="List of paths to the video files to track.",
        type=path,
        nargs="+",
    )
    add_argument(
        "intensity_ths",
        help=(
            "Blob intensity thresholds. If using background subtraction, the second value is the background difference threshold."
        ),
        type=float,
        nargs=2,
    )
    add_argument(
        "area_ths", help="Blob area thresholds (min, max).", type=float, nargs=2
    )
    add_argument(
        "tracking_intervals",
        help=(
            'Tracking intervals in frames. Examples: "0,100", "[0,100]", "[0,100] [150,200] ...". '
            "If not set, the whole video is tracked."
        ),
        type=pair_of_ints,
        nargs="+",
    )
    add_argument(
        "number_of_animals",
        help="Number of different animals that appear in the video.",
        type=int,
    )
    add_argument(
        "use_bkg",
        help="Compute and extract background to improve blob identification.",
        type=Bool,
    )
    add_argument(
        "resolution_reduction",
        help="Video resolution reduction factor for identification images. 0 = maximum reduction, 1 = no reduction.",
        type=float,
    )
    add_argument(
        "exclusive_rois",
        "Treat each separate ROI as a closed group of identities.",
        type=Bool,
    )
    add_argument(
        "track_wo_identities",
        "Track the video without assigning identities (no AI identification).",
        type=Bool,
    )
    add_argument(
        "ROI_list",
        help="List of polygons defining the Region Of Interest (ROI).",
        type=str,
        nargs="+",
    )

    # Output
    add_argument(
        "output_dir",
        help=(
            "Output directory where the session folder will be saved. Defaults to the parent directory of the video files."
        ),
        type=path,
        group="Output",
    )
    add_argument(
        "trajectories_formats",
        "List of formats for saving trajectory files. Choices: h5, npy, csv, pickle.",
        type=str,
        group="Output",
        choices=["h5", "npy", "csv", "pickle"],
        nargs="+",
    )
    add_argument(
        "bounding_box_images_in_ram",
        "If true, bounding box images are kept in RAM until no longer needed. Otherwise, they are saved to disk and loaded as needed.",
        type=Bool,
        group="Output",
    )
    add_argument(
        "DATA_POLICY",
        "Data retention policy for the session folder after successful tracking.",
        choices=[
            "trajectories",
            "validation",
            "knowledge_transfer",
            "idmatcher.ai",
            "all",
        ],
        type=str,
        group="Output",
    )

    # Background
    add_argument(
        "BACKGROUND_SUBTRACTION_STAT",
        "Statistical method to compute the background (choices: median, mean, max, min). Set a file path to load a custom background image.",
        type=str,
        group="Background Subtraction",
    )
    add_argument(
        "NUMBER_OF_FRAMES_FOR_BACKGROUND",
        "Number of frames used to compute the background. More frames increase accuracy but also memory and computation time.",
        type=int,
        group="Background Subtraction",
    )

    # Parallel processing
    add_argument(
        "number_of_parallel_workers",
        "Maximum number of parallel workers for segmentation and identification image creation. "
        "Negative: CPUs minus value. Zero: half of CPUs (max 4). One: no multiprocessing.",
        type=int,
        group="Parallel processing",
    )
    add_argument(
        "FRAMES_PER_EPISODE",
        "Maximum number of frames per video chunk (episode) for parallel processing.",
        type=int,
        group="Parallel processing",
    )

    # Knowledge transfer
    add_argument(
        "KNOWLEDGE_TRANSFER_FOLDER",
        "Path to a previous session or accumulation folder to transfer model knowledge from.",
        type=path,
        group="Knowledge and identity transfer",
    )
    add_argument(
        "identity_transfer",
        help="If true, identities from the knowledge transfer folder are transferred to the current session.",
        type=Bool,
        group="Knowledge and identity transfer",
    )
    add_argument(
        "ID_IMAGE_SIZE",
        "Size (in pixels) of the identification images used for tracking.",
        type=int,
        group="Knowledge and identity transfer",
    )

    # Checks
    add_argument(
        "check_segmentation",
        help="Abort tracking if any frame has more blobs than animals (to avoid non-animal blobs contaminating identification).",
        type=Bool,
        group="Checks",
    )

    # Contrastive
    add_argument(
        "DISABLE_CONTRASTIVE",
        "Skip contrastive learning and go directly to the accumulation protocol.",
        type=Bool,
        group="Contrastive",
    )
    add_argument(
        "CONTRASTIVE_MAX_MBYTES",
        "Maximum RAM (in megabytes) for preloading identification images during contrastive training.",
        type=float,
        group="Contrastive",
    )
    add_argument(
        "CONTRASTIVE_MIN_ACCUMULATION",
        "Minimum fraction of images that need to be accumulated for taking the contrastive step as sufficient.",
        type=float,
        group="Contrastive",
    )
    add_argument(
        "CONTRASTIVE_BATCHSIZE",
        "Number of image pairs per batch in contrastive training. Larger values require more GPU memory.",
        type=int,
        group="Contrastive",
    )
    add_argument(
        "CONTRASTIVE_SILHOUETTE_TARGET",
        "Minimum silhouette score (0 to 1) required for contrastive training to finish.",
        type=float,
        group="Contrastive",
    )
    add_argument(
        "contrastive_patience",
        "Number of training steps without improvement before early stopping contrastive training.",
        type=int,
        group="Contrastive",
    )

    # Advanced hyperparameters
    add_argument(
        "THRESHOLD_EARLY_STOP_ACCUMULATION",
        "Fraction of accumulated images needed to early stop the accumulation process.",
        type=float,
        group="Advanced hyperparameter",
    )
    add_argument(
        "MAXIMAL_IMAGES_PER_ANIMAL",
        "Maximum number of images per animal used to train the identification CNN in each accumulation step.",
        type=int,
        group="Advanced hyperparameter",
    )
    add_argument(
        "device",
        help=(
            "Device name for torch.device() (e.g., 'cpu', 'cuda', 'cuda:0', "
            "see https://pytorch.org/docs/stable/tensor_attributes.html#torch-device). "
            "Leave empty for automatic selection."
        ),
        type=str,
        group="Advanced hyperparameter",
    )
    add_argument(
        "torch_compile",
        help="If true, models are compiled with torch.compile for faster execution (may not be compatible "
        "with all devices, see https://pytorch.org/tutorials/intermediate/torch_compile_tutorial.html).",
        type=Bool,
        group="Advanced hyperparameter",
    )

    for deprecated_param in (
        "add_time_column_to_csv",
        "convert_trajectories_to_csv_and_json",
        "protocol3_action",
        "threshold_acceptable_accumulation",
        "maximum_number_of_parachute_accumulations",
        "max_ratio_of_pretrained_images",
    ):
        add_argument(
            deprecated_param,
            help=f"The parameter {deprecated_param!r} has been removed",
            type=str,
            group="Deprecated",
        )
    return parser


def get_argparser_help():
    """Used to display argument options in docs

    Returns
    -------
    str
        idtracker.ai argument parser help
    """
    return get_parser(Session.__dict__ | conf.as_dict()).format_help()


def parse_args(defaults: dict | None = None):
    parser = get_parser(defaults or (Session.__dict__ | conf.as_dict()))
    return {k: v for k, v in vars(parser.parse_args()).items() if v is not None}


if __name__ == "__main__":
    print(get_argparser_help())



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
import logging
from collections.abc import Iterable, Sequence
from contextlib import suppress
from pathlib import Path

import numpy as np
import toml
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import QEvent, QPointF, QSize, Qt, QThread, QTimer
from qtpy.QtGui import QAction  # type: ignore[reportPrivateImportUsage]
from qtpy.QtGui import (
    QCloseEvent,
    QColorConstants,
    QIcon,
    QImage,
    QKeyEvent,
    QPainter,
    QPixmap,
)
from qtpy.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QLabel,
    QSlider,
    QSpinBox,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from .. import GUIBase
from .canvas import Canvas
from .video_paths_holder import VideoPathHolder


class AsyncFrameLoader(QThread):
    precomputed_frame = Signal(int, bool, object)

    def __init__(self, holder: VideoPathHolder) -> None:
        super().__init__()
        self.holder = holder
        self.abort = False

    def precompute_frames(self, frame_indices: Iterable[int], color: bool) -> None:
        self.abort = False
        self.frame_indices = frame_indices
        self.color = color
        return super().start()

    def run(self) -> None:
        for frame_index in self.frame_indices:
            if self.abort:
                return
            try:
                frame = self.holder.read_frame(frame_index, self.color)
            except Exception as exc:
                logging.error(f"In {self.__class__.__name__}: {exc}")
            else:
                self.precomputed_frame.emit(frame_index, self.color, frame)


class VideoPlayer(QWidget):
    painting_time = Signal(QPainter, int, object)  # np.ndarray|None
    control_bar_h = 30
    is_muted: bool = False  # Whether the video player must not draw the video frames

    def __init__(self, parent: GUIBase):
        super().__init__(parent)
        self.settings = parent.settings
        self.canvas = Canvas(self)
        self.video_path_holder = VideoPathHolder()

        self.frame_preloader = AsyncFrameLoader(VideoPathHolder())
        self.frame_preloader.precomputed_frame.connect(
            self.video_path_holder.populate_cache
        )

        self.frame_slider = QSlider(Qt.Orientation.Horizontal)
        self.frame_indicator = QSpinBox()

        self.frame_slider.valueChanged.connect(self.frame_indicator.setValue)
        self.frame_slider.sliderPressed.connect(self.stop_all)

        self.frame_indicator.valueChanged.connect(self.frame_indicator_changed)
        self.frame_indicator.setKeyboardTracking(False)
        self.frame_indicator.editingFinished.connect(self.frame_indicator.clearFocus)

        self.time_indicator_widget = QLabel()
        self.play_pause_button = QToolButton()
        self.play_pause_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.play_pause_button.setShortcut(Qt.Key.Key_Space)
        self.play_pause_button.setCheckable(True)

        icon = QIcon()
        play_icon = QIcon.fromTheme("media-playback-start")
        pause_icon = QIcon.fromTheme("media-playback-pause")

        # Fallback: use a simple triangle and two bars if theme icons are missing
        if play_icon.isNull():
            play_pixmap = QPixmap(60, 60)
            play_pixmap.fill(Qt.GlobalColor.transparent)
            painter = QPainter(play_pixmap)
            painter.setBrush(Qt.GlobalColor.green)
            points = [QPointF(15, 10), QPointF(50, 30), QPointF(15, 50)]
            painter.drawPolygon(*points)
            painter.end()
            play_icon = QIcon(play_pixmap)

        if pause_icon.isNull():
            pause_pixmap = QPixmap(60, 60)
            pause_pixmap.fill(Qt.GlobalColor.transparent)
            painter = QPainter(pause_pixmap)
            painter.setBrush(Qt.GlobalColor.gray)
            painter.drawRect(15, 10, 10, 40)
            painter.drawRect(35, 10, 10, 40)
            painter.end()
            pause_icon = QIcon(pause_pixmap)

        icon.addPixmap(play_icon.pixmap(60), QIcon.Mode.Normal, QIcon.State.Off)
        icon.addPixmap(pause_icon.pixmap(60), QIcon.Mode.Normal, QIcon.State.On)

        self.play_pause_button.setIcon(icon)
        self.play_pause_button.toggled.connect(self.play_pause_clicked)
        self.time_indicator_widget.setFixedHeight(self.control_bar_h)
        self.frame_slider.setFixedHeight(self.control_bar_h)

        self.control_bar = QHBoxLayout()
        self.control_bar.addWidget(self.play_pause_button)
        self.control_bar.addWidget(self.frame_indicator)
        self.control_bar.addWidget(self.frame_slider)
        self.control_bar.addWidget(self.time_indicator_widget)

        layout = QVBoxLayout(self)
        layout.addWidget(self.canvas)
        layout.addLayout(self.control_bar)
        self.forward_loop = QTimer()
        self.backward_loop = QTimer()
        self.forward_loop.setTimerType(Qt.TimerType.PreciseTimer)
        self.backward_loop.setTimerType(Qt.TimerType.PreciseTimer)
        self.forward_loop.timeout.connect(self.forward_step)
        self.backward_loop.timeout.connect(self.backward_step)
        self.fps = 1
        self.drawn_frame = -1
        self.speed: int = 1
        self.speed_label: str = ""
        self.canvas.painting_time.connect(self.paint_video)

        menu = parent.menuBar()
        assert menu is not None
        menu = menu.addMenu("Video player")
        assert menu is not None

        self.draw_in_color = QAction("Enable color", self)
        self.draw_in_color.setCheckable(True)
        menu.addAction(self.draw_in_color)
        self.draw_in_color.toggled.connect(self.update)
        self.draw_in_color.setChecked(
            self.settings.value("video_player/draw_in_color", False, type=bool)
        )

        self.limit_framerate = QAction("Limit framerate", self)
        self.limit_framerate.setShortcut("Ctrl+L")
        self.limit_framerate.setCheckable(True)
        menu.addAction(self.limit_framerate)
        self.limit_framerate.toggled.connect(self.limit_framerate_toggled)
        self.limit_framerate.setChecked(
            self.settings.value("video_player/limit_framerate", True, type=bool)
        )

        self.reduce_cache = QAction("Reduce memory usage", self)
        self.reduce_cache.setCheckable(True)
        menu.addAction(self.reduce_cache)
        self.reduce_cache.toggled.connect(self.video_path_holder.set_cache_mode)
        self.reduce_cache.setChecked(
            self.settings.value("video_player/reduce_cache", False, type=bool)
        )

        playback_speed_action = QAction("Change playback speed", self)
        playback_speed_action.triggered.connect(lambda: ChangePlaybackSpeed(self, self.speed))  # type: ignore
        menu.addAction(playback_speed_action)

        tooltips = toml.load(Path(__file__).parent.parent / "tooltips.toml")
        self.draw_in_color.setToolTip(tooltips["color_action"])
        self.limit_framerate.setToolTip(tooltips["framerate_action"])
        self.reduce_cache.setToolTip(tooltips["reducecache_action"])
        menu.setToolTipsVisible(True)
        self.limit_framerate.setChecked(True)
        parent.installEventFilter(self)

        # We draw the video at (-0.5, -0.5) so that the pixel coordinates are in their center
        self.video_drawing_origin = QPointF(-0.5, -0.5)
        self.forward_is_pressed = False
        self.backward_is_pressed = False

        self.key_hold_delay_timer = QTimer()
        self.key_hold_delay_timer.setSingleShot(True)
        self.key_hold_delay_timer.timeout.connect(self.keyHeld)

    def limit_framerate_toggled(self, is_framerate_limited: bool):
        interval = int(1000 / self.fps) if is_framerate_limited else 0
        self.forward_loop.setInterval(interval)
        self.backward_loop.setInterval(interval)

    def preload_frames(self, start: int, end: int):
        """Preloads the frames in the video_path_holder cache"""
        color = self.draw_in_color.isChecked()
        if self.frame_preloader.isRunning():
            self.frame_preloader.abort = True
            self.frame_preloader.wait()

        with suppress(RuntimeError):
            self.frame_preloader.precompute_frames(range(start, end), color)

    def resizeEvent(self, a0):
        super().resizeEvent(a0)
        play_btn_size = self.frame_indicator.height()
        play_icon_size = int(play_btn_size * 0.6 + 0.5)
        self.play_pause_button.setFixedSize(play_btn_size, play_btn_size)
        self.play_pause_button.setIconSize(QSize(play_icon_size, play_icon_size))

    def closeEvent(self, event: QCloseEvent):
        self.settings.setValue(
            "video_player/reduce_cache", self.reduce_cache.isChecked()
        )
        if self.draw_in_color.isEnabled():
            self.settings.setValue(
                "video_player/draw_in_color", self.draw_in_color.isChecked()
            )
        self.settings.setValue(
            "video_player/limit_framerate", self.limit_framerate.isChecked()
        )
        self.video_path_holder.clear_cache()
        super().closeEvent(event)

    def event(self, event: QEvent) -> bool:
        if event.type() == QEvent.Type.WindowBlocked:
            self.stop_all()
        return super().event(event)

    def stop_all(self):
        self.play_pause_button.setChecked(False)
        self.forward_loop.stop()
        self.backward_loop.stop()

    def play_pause_clicked(self, play: bool):
        self.forward_loop.stop()
        self.backward_loop.stop()
        if play:
            self.forward_loop.start()
        else:
            self.forward_loop.stop()

    def frame_indicator_changed(self, frame_indicator_value):
        self.frame_slider.setValue(frame_indicator_value)
        self.update()

    def setCurrentFrame(self, frame, force_update=False):
        self.stop_all()
        if force_update and self.frame_indicator.value() == frame:
            self.update()
        self.frame_indicator.setValue(frame)

    @property
    def current_frame(self) -> int:
        return self.frame_indicator.value()

    @property
    def current_time(self) -> str:
        seconds = int(self.current_frame / self.fps)
        minutes = (seconds // 60) % 60
        hours = seconds // 3600
        return f"{hours:02d}:{minutes:02d}:{seconds % 60:02d}"

    def paint_video(self, painter: QPainter) -> None:
        if not self.isEnabled():
            return

        current_frame = self.current_frame
        self.time_indicator_widget.setText(self.current_time)
        color = self.draw_in_color.isChecked()

        try:
            # TODO run in a thread and display info in case it takes too long
            frame = self.video_path_holder.frame(current_frame, color)
        except RuntimeError as exc:  # unreadable frame by OpenCV
            if self.drawn_frame != current_frame:
                logging.error(exc)  # avoid printing multiple equal logs
            painter.fillRect(
                0, 0, self.video_width, self.video_height, QColorConstants.DarkGray
            )
            painter.setPen(QColorConstants.White)
            painter.drawText(
                painter.window(),
                Qt.AlignmentFlag.AlignCenter,
                str(exc).replace(" of ", " of\n"),
            )
            self.painting_time.emit(painter, current_frame, None)
        else:
            if not self.is_muted:
                painter.drawPixmap(
                    self.video_drawing_origin,
                    QPixmap.fromImage(
                        QImage(
                            frame.data,
                            frame.shape[1],
                            frame.shape[0],
                            frame.shape[1] * 3 if color else frame.shape[1],
                            (
                                QImage.Format.Format_BGR888
                                if color
                                else QImage.Format.Format_Grayscale8
                            ),
                        )
                    ),
                )
            self.painting_time.emit(painter, current_frame, frame)

        if self.speed_label:
            painter.resetTransform()
            painter.setFont(self.font())
            painter.setPen(QColorConstants.White)
            painter.drawText(
                self.canvas.rect(), Qt.AlignmentFlag.AlignBottom, self.speed_label
            )

        self.drawn_frame = current_frame

    def backward_step(self) -> None:
        new_frame = self.current_frame - self.speed
        if new_frame < 0:
            new_frame = self.n_frames
        self.frame_indicator.setValue(new_frame)

    def forward_step(self) -> None:
        new_frame = self.current_frame + self.speed
        if new_frame >= self.n_frames:
            new_frame = 0
        self.frame_indicator.setValue(new_frame)

    def eventFilter(self, object, event: QEvent) -> bool:
        """Catch key events even when VideoPlayer is not in focus."""
        if event.type() == QEvent.Type.KeyPress:
            assert isinstance(event, QKeyEvent)
            return self.keyPressEvent_from_eventFilter(event)
        if event.type() == QEvent.Type.KeyRelease:
            assert isinstance(event, QKeyEvent)
            return self.keyReleaseEvent_from_eventFilter(event)
        return False  # keep processing the event

    def keyHeld(self) -> None:
        if self.forward_is_pressed:
            self.forward_loop.start()
        if self.backward_is_pressed:
            self.backward_loop.start()

    def keyPressEvent_from_eventFilter(self, event: QKeyEvent) -> bool:
        if (
            event.isAutoRepeat()
            or not self.isEnabled()
            or event.modifiers()
            not in (Qt.KeyboardModifier.NoModifier, Qt.KeyboardModifier.KeypadModifier)
        ):
            return False
        key = event.key()
        if key in (Qt.Key.Key_D, Qt.Key.Key_Right):
            self.play_pause_button.setChecked(False)
            self.forward_step()
            self.forward_is_pressed = True
            self.key_hold_delay_timer.start(200)
            return True
        if key in (Qt.Key.Key_A, Qt.Key.Key_Left):
            self.play_pause_button.setChecked(False)
            self.backward_step()
            self.backward_is_pressed = True
            self.key_hold_delay_timer.start(200)
            return True
        with suppress(ValueError):
            self.setSpeed(int(event.text()))
            return True
        return False

    def keyReleaseEvent_from_eventFilter(self, event: QKeyEvent) -> bool:
        if event.isAutoRepeat():
            return False
        key = event.key()
        if key in (Qt.Key.Key_D, Qt.Key.Key_Right):
            self.forward_is_pressed = False
            self.forward_loop.stop()
            self.key_hold_delay_timer.stop()
            return True
        if key in (Qt.Key.Key_A, Qt.Key.Key_Left):
            self.backward_is_pressed = False
            self.backward_loop.stop()
            self.key_hold_delay_timer.stop()
            return True
        return False

    def setSpeed(self, value: int):
        if value == 0:
            return
        self.speed = 2 ** (value - 1)
        self.speed_label = f"Speed x{self.speed}" if self.speed != 1 else ""
        self.update()

    def update_video_paths(
        self,
        video_paths: Sequence[Path | str],
        n_frames: int,
        video_size: tuple[int, int],
        fps: float,
    ) -> None:
        self.fps = fps
        self.limit_framerate_toggled(self.limit_framerate.isChecked())
        self.n_frames = n_frames
        self.video_width, self.video_height = video_size
        video_paths = [Path(p) for p in video_paths]
        self.video_path_holder.load_paths(video_paths)
        self.frame_preloader.holder.load_paths(video_paths)
        self.frame_slider.setMaximum(n_frames - 1)
        self.frame_indicator.setMaximum(n_frames - 1)
        self.frame_indicator.setValue(0)
        self.canvas.adjust_zoom_to(video_size[0], video_size[1])
        self.update()

    def reorder_video_paths(self, video_paths: Sequence[Path]) -> None:
        self.video_path_holder.load_paths(video_paths)
        self.frame_preloader.holder.load_paths(video_paths)
        self.update()

    def center_canvas_at(self, x: float, y: float, zoom_scale: float) -> None:
        self.canvas.zoom = zoom_scale / min(self.width(), self.height())
        self.canvas.centerX = x
        self.canvas.centerY = y


class ChangePlaybackSpeed(QDialog):
    def __init__(self, parent: VideoPlayer, current: int) -> None:
        super().__init__(parent)
        self.setWindowFlags(Qt.WindowType.Popup)
        self.setLayout(QVBoxLayout())
        self.slider = QSlider(Qt.Orientation.Horizontal)
        layout = self.layout()
        assert layout is not None
        layout.addWidget(self.slider)
        layout.addWidget(QLabel("Change the speed by pressing a numeric key"))
        self.slider.setMinimum(1)
        self.slider.setMaximum(9)
        self.slider.setValue(int(np.log2(current)) + 1)
        self.slider.valueChanged.connect(parent.setSpeed)
        self.exec()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        with suppress(ValueError):
            self.slider.setValue(int(event.text()))



import logging
import multiprocessing
import shutil
import sys
from argparse import ArgumentError
from importlib.metadata import version
from importlib.resources import files
from pathlib import Path
from typing import Any

from idtrackerai import IdtrackeraiError, Session, conf
from idtrackerai.utils import (
    LOGGING_QUEUE,
    load_toml,
    manage_exception,
    pprint_dict,
    setup_logging_queue,
    wrap_entrypoint,
)

from .arg_parser import parse_args


def gather_input_parameters() -> tuple[bool, dict[str, Any]]:
    parameters = {}
    if Path("local_settings.py").is_file():
        logging.warning("Deprecated local_settings format found in ./local_settings.py")

    local_settings_path = Path("local_settings.toml")
    if local_settings_path.is_file():
        parameters = load_toml(local_settings_path)

    try:
        terminal_args = parse_args()
    except ArgumentError as exc:
        raise IdtrackeraiError() from exc

    ready_to_track = terminal_args.pop("track")

    if "parameters" in terminal_args:
        for parameter_file in terminal_args.pop("parameters"):
            parameters.update(load_toml(parameter_file))
    else:
        logging.info("No parameter files detected")

    if terminal_args:
        logging.info(
            pprint_dict(terminal_args, "Terminal arguments"), extra={"markup": True}
        )
        parameters.update(terminal_args)
    else:
        logging.info("No terminal arguments detected")
    return ready_to_track, parameters


@wrap_entrypoint
def main() -> bool:
    """The command `idtrackerai` runs this function"""
    ready_to_track, user_parameters = gather_input_parameters()

    session = Session()
    non_recognized_params_1 = conf.set_parameters(**user_parameters)
    non_recognized_params_2 = session.set_parameters(**user_parameters)

    non_recognized_params = non_recognized_params_1 & non_recognized_params_2

    if "add_time_column_to_csv" in non_recognized_params:
        raise IdtrackeraiError(
            'The parameter "add_time_column_to_csv" has been removed and set to always True'
        )
    if "convert_trajectories_to_csv_and_json" in non_recognized_params:
        raise IdtrackeraiError(
            'The parameter "CONVERT_TRAJECTORIES_TO_CSV_AND_JSON" has been replaced by '
            '"TRAJECTORIES_FORMATS"\nCheck the documentation https://idtracker.ai/latest/user_guide/usage.html#output'
        )

    p3_deprecated = {
        "protocol3_action",
        "threshold_acceptable_accumulation",
        "maximum_number_of_parachute_accumulations",
        "max_ratio_of_pretrained_images",
    }
    if p3_deprecated & non_recognized_params:
        raise IdtrackeraiError(
            "The following parameters have been removed from idtracker.ai "
            f"because Protocol 3 has been replaced by Contrastive Protocol:\n{p3_deprecated}"
        )

    if non_recognized_params:
        raise IdtrackeraiError(f"Not recognized parameters: {non_recognized_params}")

    if not ready_to_track:
        ready_to_track = run_segmentation_GUI(session)
        if not ready_to_track:
            return False

    from idtrackerai.base.run import RunIdTrackerAi

    return RunIdTrackerAi(session).track_video()


def run_segmentation_GUI(session: Session | None = None) -> bool:
    """Run the segmentation GUI in a separate process to catch segmentation faults"""
    # https://stackoverflow.com/a/10415215
    communication_queue = multiprocessing.Queue()

    if sys.platform.startswith("win") or sys.platform == "darwin":
        communication_dict = run_gui_in_main_process(session)
    else:
        # On Linux, we run the GUI in a separate process to catch libxcb errors
        p = multiprocessing.Process(
            target=run_gui_in_child_process,
            args=(session, communication_queue, LOGGING_QUEUE),
        )
        p.start()
        p.join()

        if p.exitcode == -6 and sys.platform == "linux":
            raise RuntimeError(
                f"QApplication crashed with exit code {p.exitcode}. This is a known issue with the "
                "Qt library on Linux. Try to run 'sudo apt install libxcb-cursor0'."
            )
        elif p.exitcode != 0:
            raise RuntimeError(f"QApplication crashed with exit code {p.exitcode}")

        communication_dict = communication_queue.get()

    run_idtrackerai = communication_dict.get("run_idtrackerai", False)
    if session is not None:
        session.__dict__.update(communication_dict)
    return run_idtrackerai


def run_gui_in_child_process(
    session: Session | None,
    communication_queue: multiprocessing.Queue,
    logging_queue: multiprocessing.Queue,
) -> None:
    if logging_queue:
        setup_logging_queue(logging_queue)

    try:
        from qtpy.QtWidgets import QApplication

        from idtrackerai.segmentation_app import SegmentationGUI
    except ImportError as exc:
        raise IdtrackeraiError(
            "\n\tRUNNING AN IDTRACKER.AI INSTALLATION WITHOUT ANY QT BINDING.\n\tGUIs"
            " are not available, only tracking directly from the terminal with the"
            " `--track` flag.\n\tRun `pip install pyqt6` or `pip install pyqt5` to"
            " build a Qt binding"
        ) from exc

    # this catches exceptions when raised inside Qt
    def excepthook(exc_type, exc_value, exc_tb):
        assert QApplication  # Pylance is happier with this
        QApplication.quit()
        manage_exception(exc_value)

    sys.excepthook = excepthook
    app = QApplication(sys.argv)

    window = SegmentationGUI(session)
    window.show()
    app.exec()

    # communicate to main process the new session parameters and if the user wants to track
    communication_queue.put(
        (session.__dict__ if session is not None else {})
        | {"run_idtrackerai": window.run_idtrackerai_after_closing}
    )


def run_gui_in_main_process(session: Session | None) -> dict[str, Any]:
    try:
        from qtpy.QtWidgets import QApplication

        from idtrackerai.segmentation_app import SegmentationGUI
    except ImportError as exc:
        raise IdtrackeraiError(
            "\n\tRUNNING AN IDTRACKER.AI INSTALLATION WITHOUT ANY QT BINDING.\n\tGUIs"
            " are not available, only tracking directly from the terminal with the"
            " `--track` flag.\n\tRun `pip install pyqt6` or `pip install pyqt5` to"
            " build a Qt binding"
        ) from exc

    # this catches exceptions when raised inside Qt
    def excepthook(exc_type, exc_value, exc_tb):
        assert QApplication  # Pylance is happier with this
        QApplication.quit()
        manage_exception(exc_value)

    sys.excepthook = excepthook
    app = QApplication(sys.argv)

    window = SegmentationGUI(session)
    window.show()
    app.exec()

    return (session.__dict__ if session is not None else {}) | {
        "run_idtrackerai": window.run_idtrackerai_after_closing
    }


@wrap_entrypoint
def general_test():
    from datetime import datetime

    logging.info("Starting general idtrackerai test")

    COMPRESSED_VIDEO_PATH = Path(str(files("idtrackerai"))) / "data" / "test_B.avi"

    video_path = Path.cwd() / COMPRESSED_VIDEO_PATH.name
    shutil.copyfile(COMPRESSED_VIDEO_PATH, video_path)

    session = Session()
    session.set_parameters(
        name="test",
        video_paths=video_path,
        tracking_intervals=None,
        intensity_ths=[0, 130],
        area_ths=[150, float("inf")],
        number_of_animals=8,
        check_segmentation=False,
        roi_list=None,
        track_wo_identities=False,
        use_bkg=False,
    )

    _ready_to_track, user_parameters = gather_input_parameters()
    non_recognized_params_1 = conf.set_parameters(**user_parameters)
    non_recognized_params_2 = session.set_parameters(**user_parameters)
    non_recognized_params = non_recognized_params_1 & non_recognized_params_2
    if non_recognized_params:
        raise IdtrackeraiError(f"Not recognized parameters: {non_recognized_params}")
    from idtrackerai.base.run import RunIdTrackerAi

    start = datetime.now()
    success = RunIdTrackerAi(session).track_video()
    if success:
        logging.info(
            "[green]Test passed successfully in %s with version %s",
            str(datetime.now() - start).split(".")[0],
            version("idtrackerai"),
            extra={"markup": True},
        )


if __name__ == "__main__":
    main()



from .animals_detection import animals_detection_API
from .segmentation import (
    generate_background_from_frame_stack,
    generate_frame_stack,
    load_custom_background,
    process_frame,
    to_gray_scale,
)

__all__ = [
    "animals_detection_API",
    "generate_background_from_frame_stack",
    "generate_frame_stack",
    "process_frame",
    "to_gray_scale",
    "load_custom_background",
]



from collections.abc import Sequence

from qtpy.QtCore import Qt, Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtWidgets import QCheckBox, QHBoxLayout, QLabel, QMessageBox, QWidget

from idtrackerai.GUI_tools import InvertibleSlider, LabelRangeSlider


class AreaThresholds(QWidget):
    valueChanged = Signal(tuple)

    def __init__(self):
        super().__init__()
        layout = QHBoxLayout()
        self.setLayout(layout)
        self.upper_limit = QCheckBox("Upper\nlimit")
        self.upper_limit.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        area_th_label = QLabel("Blob area\nthresholds")
        area_th_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.range_slider = LabelRangeSlider(
            parent=self, min=1, max=60000, block_upper=False
        )
        self.range_slider._min_label.setVisible(False)
        self.range_slider.setVisible(False)
        self.simple_slider = InvertibleSlider(1, 10000)
        self.simple_slider.set_inverted(True)
        self._upper_limit_confirmed = False  # Track if user confirmed
        self.upper_limit.toggled.connect(self.upper_limit_changed)
        layout.addWidget(area_th_label)
        layout.addWidget(self.range_slider)
        layout.addWidget(self.simple_slider)
        layout.addWidget(self.upper_limit)

        self.range_slider.valueChanged.connect(self.valueChanged.emit)
        self.simple_slider.valueChanged.connect(
            lambda val: self.valueChanged.emit((val, float("inf")))
        )

    def upper_limit_changed(self, upper_limit: bool):
        if upper_limit and not self._upper_limit_confirmed:
            reply = QMessageBox.question(
                self,
                "Enable Upper Limit?",
                "<qt>This will enable an upper limit for blob areas.<br>"
                "Large blobs from animal crossings <b>should not</b> be filtered out. "
                "Only blobs that do not correspond to animals should be excluded.<br><br>"
                "Do you want to enable the upper limit?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )
            if reply == QMessageBox.StandardButton.Yes:
                self._upper_limit_confirmed = True
            else:
                self.upper_limit.setChecked(False)
                return
        self.range_slider.setVisible(upper_limit)
        self.simple_slider.setVisible(not upper_limit)
        self.valueChanged.emit(self.value())

    def setValue(self, value: Sequence[float] | None):
        if value is None:
            value = (50, float("inf"))  # default value
        low_val = int(value[0])
        self.simple_slider.setValue(low_val)
        if value[1] == float("inf"):
            self.range_slider.setValue((low_val, self.range_slider.maximum()))
            self.upper_limit.setChecked(False)
        else:
            self.range_slider.setValue((low_val, int(value[1])))
            self.upper_limit.setChecked(True)
        self.valueChanged.emit(self.value())

    def value(self) -> tuple[float, float]:
        if self.upper_limit.isChecked():
            return self.range_slider.value()
        return (float(self.simple_slider.value()), float("inf"))



import logging
import os
from collections.abc import Callable, Iterator, Sequence
from functools import partial
from itertools import count
from pathlib import Path
from typing import Literal, Protocol

import numpy as np
import torch
from rich.console import Console
from torch.nn import CrossEntropyLoss
from torch.optim.lr_scheduler import LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms
from torchvision.datasets.folder import VisionDataset

from idtrackerai.utils import conf, load_id_images, track

from . import DEVICE, IdCNN, IdentifierBase


class DataLoaderWithLabels(Protocol):
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[tuple[torch.Tensor, torch.Tensor]]: ...


NUMBER_OF_PIN_MEMORY_USED = 0


class StopTraining:
    epochs_before_checking_stopping_conditions: int
    overfitting_counter: int
    """Number of epochs in which the network is overfitting before
    stopping the training"""

    loss_history: list[float] = []
    is_first_accumulation: bool
    epochs_limit: int
    overfitting_limit: int
    plateau_limit: float

    def __init__(
        self,
        epochs_limit: int,
        overfitting_limit: int,
        plateau_limit: float,
        is_first_accumulation: bool = False,
    ):
        self.epochs_before_checking_stopping_conditions = 10
        self.overfitting_counter = 0
        self.loss_history: list[float] = []
        self.is_first_accumulation: bool = is_first_accumulation
        self.epochs_limit = epochs_limit
        self.overfitting_limit = overfitting_limit
        self.plateau_limit = plateau_limit

    def __call__(self, train_loss: float, val_loss: float, val_acc: float) -> str:
        self.loss_history.append(val_loss)

        if self.epochs_completed > 1 and (np.isnan(train_loss) or np.isnan(val_loss)):
            raise RuntimeError(
                f"The model diverged {train_loss=} {val_loss=}. Check the"
                " hyperparameters and the architecture of the network."
            )

        # check if it did not reached the epochs limit
        if self.epochs_completed >= self.epochs_limit:
            return (
                "The number of epochs completed is larger than the number "
                "of epochs set for training, we stop the training"
            )

        if self.epochs_completed <= self.epochs_before_checking_stopping_conditions:
            return ""

        # check that the model is not overfitting or if it reached
        # a stable saddle (minimum)
        loss_trend = np.nanmean(
            self.loss_history[-self.epochs_before_checking_stopping_conditions : -1]
        )

        # The validation loss in the first 10 epochs could have exploded
        # but being decreasing.
        if np.isnan(loss_trend):
            loss_trend = float("inf")
        losses_difference = float(loss_trend) - val_loss

        # check overfitting
        if losses_difference < 0.0:
            self.overfitting_counter += 1
            if self.overfitting_counter >= self.overfitting_limit:
                return "Overfitting"
        else:
            self.overfitting_counter = 0

        # check if the error is not decreasing much

        if abs(losses_difference) < self.plateau_limit * val_loss:
            return "The losses difference is very small, we stop the training"

        # if the individual accuracies in validation are 1. for all the animals
        if val_acc == 1.0:
            return (
                "The individual accuracies in validation is 100%, we stop the training"
            )

        # if the validation loss is 0.
        if loss_trend == 0.0 or val_loss == 0.0:
            return "The validation loss is 0.0, we stop the training"

        return ""

    @property
    def epochs_completed(self):
        return len(self.loss_history)


def train_loop(
    model: IdCNN,
    criterion: CrossEntropyLoss,
    optimizer: Optimizer,
    train_loader: DataLoaderWithLabels,
    val_loader: DataLoaderWithLabels,
    stop_training: Callable[[float, float, float], str],
    scheduler: "LRScheduler | None" = None,
):
    logging.debug("Entering the training loop...")
    with Console().status("[red]Initializing training...") as status:
        for epoch in count(1):
            train_loss = train(train_loader, model, criterion, optimizer, scheduler)
            val_loss, val_acc = evaluate(val_loader, model, criterion)

            status.update(
                f"[red]Epoch {epoch:2}: training_loss={train_loss:.5f}, "
                f"validation_loss={val_loss:.5f} and accuracy={val_acc:.3%}"
            )
            stop_message = stop_training(train_loss, val_loss, val_acc)
            if stop_message:
                break
        else:
            raise

    logging.info(stop_message)
    logging.info("Last epoch: %s", status.status, extra={"markup": True})
    logging.info("Network trained")


def train(
    train_loader: DataLoaderWithLabels,
    model: IdCNN,
    criterion: CrossEntropyLoss,
    optimizer: Optimizer,
    scheduler: "LRScheduler | None" = None,
) -> float:
    """Trains trains a network using a learner, a given train_loader"""
    losses = 0
    n_predictions = 0

    model.train()

    for images, labels in train_loader:
        images = images.to(DEVICE, non_blocking=True)
        labels = labels.to(DEVICE, non_blocking=True)

        out = model(images)
        loss = criterion(out, labels)

        optimizer.zero_grad(set_to_none=True)
        loss.backward()
        optimizer.step()

        losses += loss.item() * len(images)
        n_predictions += len(images)

    if scheduler is not None:
        scheduler.step()
    return losses / n_predictions


@torch.inference_mode()
def evaluate(
    eval_loader: DataLoaderWithLabels, model: IdCNN, criterion: CrossEntropyLoss
) -> tuple[float, float]:
    losses = 0
    n_predictions = 0
    n_right_guess = 0

    model.eval()

    for images, labels in eval_loader:
        images = images.to(DEVICE, non_blocking=True)
        labels = labels.to(DEVICE, non_blocking=True)

        out = model(images)
        loss = criterion(out, labels)
        n_predictions += len(labels)
        n_right_guess += (out.max(1).indices == labels).count_nonzero().item()

        losses += loss.item() * len(images)

    return losses / n_predictions, n_right_guess / n_predictions


@torch.inference_mode()
def evaluate_only_acc(eval_loader: DataLoaderWithLabels, model: IdCNN) -> float:
    model.eval()
    n_predictions = 0
    n_right_guess = 0

    for images, labels in eval_loader:
        images = images.to(DEVICE, non_blocking=True)
        labels = labels.to(DEVICE, non_blocking=True)

        predictions = model(images).max(1).indices
        n_predictions += len(labels)
        n_right_guess += (predictions == labels).count_nonzero().item()

    return n_right_guess / n_predictions


class ImageDataset(VisionDataset):
    def __init__(self, images: np.ndarray, labels: np.ndarray | None, transform=None):
        super().__init__("", transform=transform)

        if images.ndim <= 3:
            images = np.expand_dims(images, axis=-1)

        self.images = images
        self.labels = (
            labels.astype(np.int64)
            if labels is not None
            else np.zeros(len(images), np.int64)
        )
        assert len(self.images) == len(self.labels)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        image = self.images[index]
        target = self.labels[index]
        if self.transform is not None:
            image = self.transform(image)
        return image, target


def get_dataloader(
    scope: Literal["training", "validation", "test"],
    images: np.ndarray,
    labels: np.ndarray | None = None,
    batch_size: int = conf.BATCH_SIZE_PREDICTIONS,
) -> DataLoaderWithLabels:
    global NUMBER_OF_PIN_MEMORY_USED
    logging.info(
        "Creating %s dataloader with %d images"
        + (
            f" labeled with {len(np.unique(labels))} distinct classes"
            if labels is not None
            else ""
        ),
        scope,
        len(images),
    )

    if scope == "training":
        assert labels is not None
        rotated_images = np.rot90(images, 2, axes=(1, 2))
        images = np.concatenate([images, rotated_images])
        labels = np.concatenate([labels, labels])

    # We set pin_memory on training only because of https://github.com/pytorch/pytorch/issues/91252
    # And we limit the number of dataloaders created with pin_memory
    pin_memory = scope == "training"
    if NUMBER_OF_PIN_MEMORY_USED > 5:
        pin_memory = False
    if pin_memory:
        NUMBER_OF_PIN_MEMORY_USED += 1
    # logging.debug(f"{pin_memory=}")

    return DataLoader(
        ImageDataset(images, labels, transforms.ToTensor()),
        batch_size=batch_size,
        shuffle=scope == "training",
        num_workers=1 if os.name == "nt" else 4,  # windows
        persistent_workers=True,
        pin_memory=pin_memory,
    )


@torch.inference_mode()
def get_predictions(
    model: IdentifierBase,
    image_location: Sequence[tuple[int, int]] | np.ndarray,
    id_images_paths: Sequence[Path],
    kind: str = "identities",
) -> tuple[np.ndarray, np.ndarray]:
    logging.debug("Predicting %s of %d images", kind, len(image_location), stacklevel=2)
    predictions = np.empty(len(image_location), np.int32)
    probabilities = np.empty(len(image_location), np.float32)
    index = 0
    model.to(DEVICE)
    model.eval()
    dataloader = get_onthefly_dataloader(image_location, id_images_paths)
    for images, _labels in track(dataloader, "Predicting " + kind):
        classification_probs = model(images.to(DEVICE))
        batch_probabilities, batch_predictions = classification_probs.max(dim=1)
        batch_predictions += 1
        batch_size = len(batch_predictions)
        predictions[index : index + batch_size] = batch_predictions.numpy(force=True)
        probabilities[index : index + batch_size] = batch_probabilities.numpy(
            force=True
        )
        index += batch_size
    assert index == len(predictions) == len(probabilities)
    return predictions, probabilities


def get_onthefly_dataloader(
    image_locations: Sequence[tuple[int, int]] | np.ndarray,
    id_images_paths: Sequence[Path],
    labels: Sequence | np.ndarray | None = None,
) -> DataLoaderWithLabels:
    """This dataloader will load images from disk "on the fly" when asked in
    every batch. It is fast due to PyTorch parallelization with `num_workers`
    and it is very RAM efficient. Only recommended to use in predictions.
    For training it is best to use preloaded images."""
    logging.info("Creating on-the-fly DataLoader with %d images", len(image_locations))
    return DataLoader(
        SimpleDataset(image_locations, labels),
        conf.BATCH_SIZE_PREDICTIONS,
        num_workers=2,
        collate_fn=partial(collate_fun, id_images_paths=id_images_paths),
        # pin_memory=True, https://github.com/pytorch/pytorch/issues/91252
    )


def collate_fun(
    locations_and_labels: list[tuple[tuple[int, int], int]],
    id_images_paths: Sequence[Path],
) -> tuple[torch.Tensor, torch.Tensor]:
    """Receives the batch images locations (episode and index).
    These are used to load the images and generate the batch tensor"""
    locations, labels = zip(*locations_and_labels)
    return (
        torch.from_numpy(
            load_id_images(id_images_paths, locations, verbose=False, dtype=np.float32)
        ).unsqueeze(1)
        / 255,
        torch.tensor(labels),
    )


class SimpleDataset(Dataset):
    def __init__(
        self, images: Sequence | np.ndarray, labels: Sequence | np.ndarray | None = None
    ):
        super().__init__()
        self.images = images
        if labels is not None:
            self.labels = np.asarray(labels).astype(np.int64)
        else:
            self.labels = np.full(len(images), -1, np.int64)
        assert self.labels.ndim == 1
        assert len(self.images) == len(self.labels)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index: int):
        return self.images[index], self.labels[index]



import json
import logging
from collections.abc import Callable, Sequence
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from functools import wraps
from math import sqrt
from pathlib import Path
from shutil import rmtree
from typing import IO

import cv2
import h5py
import numpy as np
import toml
from deprecated.sphinx import deprecated as _deprecated

from .rich_utils import track


def deprecated(version: str = "", reason: str = "", **kwargs):
    """Decorator to set the __doc__ of a method to its __name__ and mark it as deprecated.
    It helps Sphinx to render the documentation correctly."""

    def decorator(func: Callable):
        if func.__doc__ is None:
            func.__doc__ = func.__name__.replace("_", " ").capitalize()

        @_deprecated(version=version, reason=reason, **kwargs)
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator


class IdtrackeraiError(Exception):
    """Error related to idtracker.ai. Used to communicate a user-friendly error
    message without tracebacks nor advanced software concepts.

    Works together with :func:`idtrackerai.utils.logging_utils.manage_exception`"""

    def __str__(self) -> str:
        # add __cause__ to string representation
        out = super().__str__()
        if self.__cause__ is None:
            return out
        else:
            return (
                f"{out}\n    Internal error message: {self.__cause__} ({self.__cause__.__class__.__name__})"
                if out
                else str(self.__cause__)
            )


def delete_attributes_from_object(object_to_modify, list_of_attributes):
    for attribute in list_of_attributes:
        if hasattr(object_to_modify, attribute):
            delattr(object_to_modify, attribute)


def load_toml(path: Path, name: str | None = None) -> dict:
    # we do not check if file exists, pathlib will do it for us

    # Avoid loading huge video files loaded by mistake in CLI with "--load"
    if path.stat().st_size > 5000000:
        raise IdtrackeraiError(
            f"{path} takes {path.stat().st_size / (1024**2):.1f} MB, it does not seem"
            " like a .toml file"
        )

    try:
        toml_dict = {
            key.lower(): value
            for key, value in toml.load(path.open(encoding="utf-8")).items()
        }

        for key, value in toml_dict.items():
            if value == "":
                toml_dict[key] = None

        logging.info(
            pprint_dict(toml_dict, name or f"Loading parameters from {path}"),
            extra={"markup": True},
        )
        return toml_dict
    except Exception as exc:
        raise IdtrackeraiError(f"Could not read toml file {path}") from exc


def create_dir(path: Path, remove_existing=False):
    if path.is_dir():
        if remove_existing:
            rmtree(path, ignore_errors=True)
            path.mkdir()
            logging.info(f"Directory {path} has been emptied", stacklevel=2)
        else:
            logging.info(f"Directory {path} already exists", stacklevel=2)
    else:
        if not path.parent.is_dir():
            path.parent.mkdir()
        path.mkdir()
        logging.info(f"Directory {path} has been created", stacklevel=2)


def remove_dir(path: Path):
    if path.is_dir():
        rmtree(path, ignore_errors=True)
        logging.info(f"Directory {path} has been removed", stacklevel=2)
    else:
        logging.info(f"Directory {path} not found, can't remove", stacklevel=2)


def remove_file(path: Path):
    if path.is_file():
        path.unlink()
        logging.info(f"File {path} has been removed", stacklevel=2)


def assert_all_files_exist(paths: list[Path]):
    """Returns FileNotFoundError if any of the paths is not an existing file"""
    for path in paths:
        if not path.is_file():
            raise FileNotFoundError(f"File {path} not found")


def get_vertices_from_label(label: str, close=False) -> np.ndarray:
    """Transforms a string representation of a polygon from the
    ROI widget (idtrackerai_app) into a vertices np.array"""
    try:
        data = json.loads(label[10:].replace("'", '"'))
    except ValueError as exc:
        raise IdtrackeraiError(f"Not recognized ROI representation: {label!r}") from exc

    if label[2:9] == "Polygon":
        vertices = np.asarray(data)
    elif label[2:9] == "Ellipse":
        vertices = np.asarray(
            cv2.ellipse2Poly(data["center"], data["axes"], data["angle"], 0, 360, 2)
        )
    else:
        raise TypeError(label)

    if close:
        return np.vstack([vertices, vertices[0]])
    return vertices


def build_ROI_mask_from_list(
    list_of_ROIs: None | list[str] | str, width: int, height: int
) -> np.ndarray | None:
    """Transforms a list of polygons (as type str) from
    ROI widget (idtrackerai_app) into a boolean np.array mask"""
    if list_of_ROIs is None:
        return None

    ROI_mask = np.zeros((height, width), np.uint8)

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = [list_of_ROIs]

    for line in list_of_ROIs:
        vertices = (get_vertices_from_label(line) + 0.5).astype(np.int32)
        if line[0] == "+":
            cv2.fillPoly(ROI_mask, (vertices,), color=[255])
        elif line[0] == "-":
            cv2.fillPoly(ROI_mask, (vertices,), color=[0])
        else:
            raise TypeError
    return ROI_mask


@dataclass(slots=True)
class Episode:
    index: int
    local_start: int
    local_end: int
    video_path: Path
    global_start: int
    global_end: int
    bbox_images: Path | None | IO[bytes] = None

    @property
    def length(self) -> int:
        return self.global_end - self.global_start


class Timer:
    """Simple class for measuring execution time during the whole process"""

    start_time: datetime | None = None
    finish_time: datetime | None = None

    def __enter__(self) -> "Timer":
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if exc_type is None:
            self.finish()

    def __init__(self, name: str = "") -> None:
        self.name = name

    def reset(self) -> None:
        self.start_time = None
        self.finish_time = None

    @property
    def interval(self) -> None | timedelta:
        if self.finish_time is None or self.start_time is None:
            return None
        return self.finish_time - self.start_time

    @property
    def started(self) -> bool:
        return self.start_time is not None

    @property
    def finished(self) -> bool:
        return self.interval is not None

    def start(self) -> None:
        logging.info(
            "[blue bold]START %s", self.name, extra={"markup": True}, stacklevel=2
        )
        self.start_time = datetime.now()

    def finish(self, raise_if_not_started=True) -> None:
        if not self.started and raise_if_not_started:
            raise RuntimeError("Timer finish method called before start method")

        self.finish_time = datetime.now()

        logging.info(
            f"[blue bold]FINISH {self.name}, it took {self}",
            extra={"markup": True},
            stacklevel=2,
        )

    def __str__(self) -> str:
        return str(self.interval or "Not finished").split(".")[0]

    @classmethod
    def from_dict(cls, d: dict) -> "Timer":
        obj = cls.__new__(cls)
        obj.name = d["name"]
        d.pop("py/object", None)

        if "interval" in d:  # v5.1.0 compatibility
            if d["start_time"] > 0:
                obj.start_time = datetime.fromtimestamp(d["start_time"])

            if d["interval"] > 0:
                obj.finish_time = datetime.fromtimestamp(
                    d["start_time"] + d["interval"]
                )

        else:
            if "start_time" in d:
                obj.start_time = datetime.fromisoformat(d["start_time"])
            if "finish_time" in d:
                obj.finish_time = datetime.fromisoformat(d["finish_time"])

        return obj


@dataclass(slots=True)
class LengthCalibration:
    "Length calibration used in the Validator to transform pixel units to user defined units."

    color: int = 0x000000
    point_A: Sequence[float] | None = None
    point_B: Sequence[float] | None = None
    distance: float | None = None

    @classmethod
    def from_dict(cls, d: dict):
        obj = cls.__new__(cls)
        obj.point_A = d.get("point_A")
        obj.point_B = d.get("point_B")
        obj.distance = d.get("distance")
        return obj

    def value(self) -> float | None:
        if self.point_A is None or self.point_B is None or self.distance is None:
            return None
        return (
            sqrt(
                (self.point_A[0] - self.point_B[0]) ** 2
                + (self.point_A[1] - self.point_B[1]) ** 2
            )
            / self.distance
        )

    def add_point(self, point: Sequence[float]) -> None:
        if self.point_A is None:
            self.point_A = point
            return
        if self.point_B is None:
            self.point_B = point
            return
        raise ValueError("Calibration is already full")

    def completed(self) -> bool:
        return self.has_two_points() and self.distance is not None

    def has_two_points(self) -> bool:
        return self.point_A is not None and self.point_B is not None

    def __str__(self) -> str:
        return f"[{self.point_A}, {self.point_B}]: {self.distance}"


def assert_knowledge_transfer_is_possible(
    knowledge_transfer_folder: Path | None, n_animals: int
) -> tuple[list[int], float | None]:
    if knowledge_transfer_folder is None:
        raise IdtrackeraiError(
            "To perform knowledge/identity transfer you "
            "need to provide a path for the variable "
            "'KNOWLEDGE_TRANSFER_FOLDER'"
        )

    model_params_path = knowledge_transfer_folder / "model_params.json"
    if model_params_path.is_file():
        model_params_dict = json.load(model_params_path.open())
        n_classes, image_size, res_reduct = get_parameters_from_model_json(
            model_params_dict
        )

    elif model_params_path.with_suffix(".npy").is_file():
        model_params_dict = np.load(
            model_params_path.with_suffix(".npy"), allow_pickle=True
        ).item()  # loading from v4
        n_classes, image_size, res_reduct = get_parameters_from_model_json(
            model_params_dict
        )

    else:
        logging.warning('"%s" file not found', model_params_path)
        n_classes, image_size = get_parameters_from_model_state_dict(
            knowledge_transfer_folder
        )
        res_reduct = None

    if n_animals != n_classes:
        raise IdtrackeraiError(
            "Tracking with knowledge/identity transfer is not possible. "
            "The number of animals in the video needs to be the same as "
            "the number of animals in the transferred network."
        )

    logging.info(
        "Tracking with knowledge transfer. "
        "The identification image size will be matched "
        f"to the image_size of the transferred network {image_size}"
        + f" as well as the resolution reduction ({res_reduct})"
        if res_reduct
        else ""
    )
    return image_size, res_reduct


def get_parameters_from_model_json(model_parameters: dict):
    image_size = model_parameters.get("image_size", [])
    n_classes = model_parameters.get(  # 5.1.6 compatibility
        "n_classes" if "n_classes" in model_parameters else "number_of_classes"
    )
    res_reduct = model_parameters.get("resolution_reduction")
    return n_classes, image_size, res_reduct


def get_parameters_from_model_state_dict(
    knowledge_transfer_folder: Path,
) -> tuple[int, list[int]]:
    logging.info("Extracting model parameters from state dictionary")
    # this import is here (not at the top of the file) to avoid its loading process
    # when loading GUIs without identity_transfer (almost always)
    import torch

    model_dict_path = knowledge_transfer_folder / "identification_network.model.pth"
    # model_dict_path can contain metadata in previous versions of idtrackerai, that's why weights_only=False
    model_state_dict: dict[str, torch.Tensor] = torch.load(
        model_dict_path, weights_only=False
    )
    if "fc2.weight" in model_state_dict:
        layer_in_dimension = model_state_dict["fc1.weight"].size(1)
        n_classes = len(model_state_dict["fc2.weight"])
    else:
        layer_in_dimension = model_state_dict["layers.9.weight"].size(1)
        n_classes = len(model_state_dict["layers.11.weight"])
    image_size = int(4 * sqrt(layer_in_dimension / 100)) + 2
    return n_classes, [image_size, image_size, 1]


def pprint_dict(d: dict, name: str = "") -> str:
    text = f"[bold blue]{name}[/]:" if name else ""

    pad = min(max(map(len, d.keys())), 30)

    for key, value in d.items():
        if isinstance(value, tuple):
            value = list(value)
        if isinstance(value, list) and value and isinstance(value[0], Path):
            value = list(map(str, value))
        if isinstance(value, Path):
            value = str(value)
        if len(repr(value)) < 50 or not isinstance(value, list):
            text += f"\n[bold]{key:>{pad}}[/]={repr(value)}"
        else:
            s = f"[{repr(value[0])}"
            for item in value[1:]:
                s += f",\n{' ' * pad}  {repr(item)}"
            s += "]"
            text += f"\n[bold]{key:>{pad}}[/]={s}"
    return text


class H5DatasetProxy:
    """Proxy class to allow pickling a h5py.Dataset with parallel Dataloaders."""

    def __init__(self, filename: str | Path, dataset_name: str):
        self.filename = filename
        self.dataset_name = dataset_name
        self.dataset: h5py.Dataset | None = None

    def __getitem__(self, item):
        if self.dataset is None:
            self.dataset = h5py.File(self.filename, "r")[self.dataset_name]  # type: ignore
        return self.dataset[item]  # type: ignore

    def __getstate__(self):
        return (self.filename, self.dataset_name)

    def __setstate__(self, state):
        self.filename, self.dataset_name = state
        self.dataset = None


def load_id_images(
    images_sources: Sequence[Path | str | h5py.Dataset | np.ndarray | H5DatasetProxy],
    images_indices: Sequence[tuple[int, int]] | np.ndarray,
    verbose=True,
    dtype: type[np.number] | None = None,
) -> np.ndarray:
    """Loads the identification images from disk.

    Parameters
    ----------
    images_sources : Sequence[Path | str | h5py.Dataset | np.ndarray]
        List of paths or datasets where the images are stored.
    images_indices : Sequence[tuple[int, int]] | np.ndarray
        List of tuples (image_index, episode) that indicate each of the images to be loaded.
    verbose : bool, optional
        If True, display a progress bar during loading, by default True.
    dtype : type[np.number] | None, optional
        Desired data type of the output array. If None the datatype from the source is used. By default None.

    Returns
    -------
    np.ndarray
        Numpy array of shape [number of images, width, height] containing the loaded images.
    """
    img_indices, episodes = np.asarray(images_indices).T

    images = None

    for episode in track(
        np.unique(episodes),
        f"Loading {len(img_indices)} identification images from disk",
        verbose=verbose,
    ):
        where = episodes == episode
        indices = img_indices[where]

        images_source = images_sources[episode]
        if isinstance(images_source, (h5py.Dataset, np.ndarray, H5DatasetProxy)):
            dataset = images_source
            file = None
        else:
            file = h5py.File(images_source, "r")
            dataset = file["id_images"]

        if (
            isinstance(dataset, np.ndarray)
            or len(indices) > 100
            or len(indices) > len(np.unique(indices))
        ):
            # for more than 100 images, it's more efficient to load the entire file and select the desired indices
            # repetitions in indices is not acceptable for specific indices reading in h5py
            # Preloaded Numpy arrays are also treated here
            episode_imgs: np.ndarray = dataset[:][indices]  # type: ignore
        else:
            # for less than 100 images, it's faster to get only the specific indices but h5py requires the indices to be sorted
            order = np.argsort(indices)
            unorder = np.empty_like(order)
            unorder[order] = np.arange(len(order))
            episode_imgs: np.ndarray = dataset[indices[order]][unorder]  # type: ignore

        if file is not None:
            file.close()  # close the file if it was opened

        if images is None:
            # We take the first iteration to extract the image shape and dtype
            images = np.empty(
                (len(images_indices), *episode_imgs.shape[1:]),
                dtype or episode_imgs.dtype,
            )

        images[where] = episode_imgs

    if images is None:
        # in case of not having any iteration
        return np.zeros((0, 30, 30), np.uint8)

    return images


def json_default(obj):
    """Encodes non JSON serializable object as dicts"""
    match obj:
        case Path():
            return str(obj)
        case LengthCalibration():
            out = asdict(obj)
            out.pop("color")
            return out
        case Timer():
            return vars(obj)
        case np.integer():
            return int(obj)
        case np.floating():
            return float(obj)
        case np.ndarray():
            return obj.tolist()
        case set():
            return list(obj)
        case datetime():
            return obj.isoformat()
        case _:
            raise ValueError(f"Could not JSON serialize {obj} of type {type(obj)}")


def json_object_hook(d: dict):
    """DEPRECATED. Decodes dicts from `json_default`"""
    py_object = d.pop("py/object", None)
    if py_object is None:
        return d
    if py_object == "Path":
        return Path(d["path"])
    if py_object == "Episode":
        return Episode(**d)
    if py_object == "Timer":
        return Timer.from_dict(d)
    if py_object == "np.ndarray":
        return np.asarray(d["values"])
    if py_object == "set":
        return set(d["values"])
    raise ValueError(f"Could not read {d} of type {py_object}")


def resolve_path(path: Path | str) -> Path:
    return Path(path).expanduser().resolve()


def clean_attrs(obj: object):
    """Removes instances attributes if they are redundant
    with the class attributes"""
    class_attr = obj.__class__.__dict__

    attributes_to_remove: list[str] = [
        attr
        for attr, value in obj.__dict__.items()
        if attr in class_attr
        and isinstance(class_attr[attr], type(value))
        and class_attr[attr] == value
    ]

    for attr in attributes_to_remove:
        delattr(obj, attr)



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import cv2
import toml
from qtpy.QtCore import Qt, QTimer
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QSpinBox,
    QSplitter,
    QVBoxLayout,
    QWidget,
)

from idtrackerai import Session
from idtrackerai.GUI_tools import GUIBase, QHLine, VideoPlayer
from idtrackerai.utils import pprint_dict

from .widgets import (
    AreaThresholds,
    BkgWidget,
    BlobInfoWidget,
    FrameAnalyzer,
    IntensityThresholds,
    OpenVideoWidget,
    ROIWidget,
    TrackingIntervalsWidget,
)


class SegmentationGUI(GUIBase):
    """The main class implementing the Segmentation app.
    From here, all widgets are coordinated and connected.

    Parameters
    ----------
    GUIBase : QMainWindow
        Base configuration for all idtracker.ai GUIs
    """

    run_idtrackerai_after_closing: bool = False

    def __init__(self, session: Session | None = None):
        super().__init__()

        self.setWindowTitle("Segmentation App")
        self.session = session or Session()
        self.documentation_url = (
            "https://idtracker.ai/latest/user_guide/segmentation_app.html"
        )

        self.open_widget = OpenVideoWidget(self, self.session.frames_per_episode)
        self.videoPlayer = VideoPlayer(self)
        self.frame_analyzer = FrameAnalyzer()
        self.blobInfo = BlobInfoWidget()
        self.bkg_widget = BkgWidget()
        self.ROI_Widget = ROIWidget(self)
        self.tracking_interval = TrackingIntervalsWidget(self)
        self.widgets_to_close.append(self.videoPlayer)

        self.intensity_thresholds = IntensityThresholds(self, min=0, max=255)
        self.area_thresholds = AreaThresholds()

        self.save_parameters = QPushButton("Save parameters")
        self.save_parameters.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.save_parameters.setShortcut("Ctrl+S")

        self.check_segm = QCheckBox("Stop tracking if #blobs > #animals")
        self.track_wo_id = QCheckBox("Track without identities")

        n_animals_row = QHBoxLayout()
        n_animals_label = QLabel("Number of animals")
        n_animals_row.addWidget(n_animals_label)
        self.n_animals = QSpinBox()
        self.n_animals.setMaximum(1000)
        self.n_animals.setMinimum(0)
        self.n_animals.setSpecialValueText("unknown")
        n_animals_row.addWidget(self.n_animals)
        n_animals_row.setAlignment(Qt.AlignmentFlag.AlignLeft)

        session_row = QHBoxLayout()
        session_label = QLabel("Session name")
        session_row.addWidget(session_label)
        self.session_name = SessionName()
        session_row.addWidget(self.session_name)
        session_row.addWidget(self.save_parameters)

        self.close_and_track_btn = QPushButton("Close window and track video")
        self.close_and_track_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Connecting widgets
        self.open_widget.path_clicked.connect(self.videoPlayer.setCurrentFrame)
        self.open_widget.new_video_paths.connect(self.new_video_paths)
        self.open_widget.new_episodes.connect(self.bkg_widget.set_new_video_paths)
        self.open_widget.new_parameters.connect(self.new_parameters)
        self.open_widget.video_paths_reordered.connect(
            self.videoPlayer.reorder_video_paths
        )
        self.open_widget.video_paths_reordered.connect(
            lambda paths: self.session_name.setPlaceholderText(
                "&".join(Path(path).stem for path in paths)
            )
        )
        self.n_animals.editingFinished.connect(self.n_animals.clearFocus)
        self.n_animals.valueChanged.connect(self.blobInfo.setNAnimals)
        self.tracking_interval.range_slider.single_value_changed.connect(
            self.videoPlayer.setCurrentFrame
        )
        self.tracking_interval.newValue.connect(self.open_widget.set_tracking_interval)
        self.tracking_interval.newValue.connect(self.blobInfo.setTrackingIntervals)
        self.intensity_thresholds.newValue.connect(
            self.frame_analyzer.set_intensity_ths
        )
        self.session_name.editingFinished.connect(self.session_name.clearFocus)
        self.save_parameters.clicked.connect(self.save_parameters_func)
        self.area_thresholds.valueChanged.connect(self.frame_analyzer.set_area_ths)
        self.close_and_track_btn.clicked.connect(self.close_and_track_video)
        self.ROI_Widget.valueChanged.connect(self.frame_analyzer.set_ROI_mask)
        self.ROI_Widget.needToDraw.connect(self.videoPlayer.update)
        self.bkg_widget.new_bkg_data.connect(self.frame_analyzer.set_bkg)
        self.bkg_widget.new_bkg_data.connect(self.intensity_thresholds.bkg_changed)
        self.frame_analyzer.new_areas.connect(self.blobInfo.setAreas)
        self.frame_analyzer.new_parameters.connect(self.videoPlayer.update)
        self.videoPlayer.painting_time.connect(self.frame_analyzer.paint_on_canvas)
        self.videoPlayer.painting_time.connect(self.ROI_Widget.paint_on_canvas)
        self.videoPlayer.canvas.click_event.connect(self.ROI_Widget.click_event)

        # Tooltips texts
        tooltips = toml.load(Path(__file__).parent / "tooltips.toml")

        self.open_widget.button_open.setToolTip(tooltips["open_btn"])
        self.open_widget.single_file_label.setToolTip(tooltips["open_path_label"])
        self.open_widget.list_of_files.setToolTip(tooltips["open_path_list"])
        self.tracking_interval.setToolTip(tooltips["tracking_interval"])
        self.ROI_Widget.setToolTip(tooltips["region_of_interest"])
        self.ROI_Widget.exclusive_rois.setToolTip(tooltips["exclusive_rois"])
        self.bkg_widget.setToolTip(tooltips["background_subtraction"])
        self.bkg_widget.bkg_stat.setToolTip(tooltips["background_stat"])
        self.bkg_widget.view_bkg.setToolTip(tooltips["background_view"])
        self.n_animals.setToolTip(tooltips["number_of_animals"])
        n_animals_label.setToolTip(tooltips["number_of_animals"])
        self.check_segm.setToolTip(tooltips["check_segm"])
        self.area_thresholds.setToolTip(tooltips["area_thresholds"])
        self.track_wo_id.setToolTip(tooltips["track_wo_id"])
        self.save_parameters.setToolTip(tooltips["save_params"])
        self.close_and_track_btn.setToolTip(tooltips["close_and_track"])
        self.blobInfo.setToolTip(tooltips["blobs_info"])
        self.session_name.setToolTip(tooltips["session_name"])
        session_label.setToolTip(tooltips["session_name"])
        self.intensity_thresholds.setToolTips(
            tooltips["intensity_thresholds_nobkg"],
            tooltips["intensity_thresholds_yesbkg"],
        )

        left_layout = QVBoxLayout()
        self.open_widget.layout().setContentsMargins(0, 8, 0, 0)
        left_layout.addWidget(self.open_widget)
        for widget in (
            QHLine(),
            self.tracking_interval,
            self.ROI_Widget,
            QHLine(),
            n_animals_row,
            self.bkg_widget,
            self.intensity_thresholds,
            self.area_thresholds,
            self.blobInfo,
            QHLine(),
        ):
            if isinstance(widget, (QVBoxLayout, QHBoxLayout)):
                widget.setContentsMargins(0, 0, 0, 0)
                superwidget = QWidget()
                superwidget.setLayout(widget)
                left_layout.addWidget(
                    superwidget, alignment=Qt.AlignmentFlag.AlignVCenter
                )
            else:
                lay = widget.layout()
                if lay is not None:
                    lay.setContentsMargins(0, 0, 0, 0)
                left_layout.addWidget(
                    widget,
                    alignment=Qt.AlignmentFlag.AlignVCenter,
                    stretch=1 if isinstance(widget, QHLine) else 3,
                )
        left_layout.addWidget(self.check_segm)
        left_layout.addWidget(self.track_wo_id)
        left_layout.addLayout(session_row)
        left_layout.addWidget(self.close_and_track_btn)

        self.videoPlayer.layout().setContentsMargins(8, 0, 0, 0)

        left = QWidget()
        left.setLayout(left_layout)

        main_splitter = QSplitter(Qt.Orientation.Horizontal)
        main_splitter.addWidget(left)
        main_splitter.addWidget(self.videoPlayer)
        main_splitter.setSizes([400, 600])
        self.centralWidget().layout().addWidget(main_splitter)
        self.list_of_widgets = self.get_list_of_widgets(left_layout)
        for widget in self.list_of_widgets:
            widget.setEnabled(False)
        self.videoPlayer.setEnabled(False)
        self.enabled = False
        self.open_widget.setEnabled(True)

        self.setTabOrder(self.n_animals, self.videoPlayer.canvas)
        self.setTabOrder(self.videoPlayer.canvas, self.n_animals)
        for widget in self.findChildren(QCheckBox):
            assert isinstance(widget, QWidget)
            widget.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        QTimer.singleShot(0, self.load_parameters)

    def manageDropedPaths(self, paths: Sequence[str]) -> None:
        self.open_widget.process_paths(paths)

    def load_parameters(self):
        """Sets all widgets to the values indicated by self.session"""
        self.open_widget.open_video_paths(self.session.video_paths)
        self.tracking_interval.setValue(self.session.tracking_intervals)
        self.ROI_Widget.setValue(self.session.roi_list, self.session.exclusive_rois)
        self.intensity_thresholds.setValue(self.session.intensity_ths)
        self.area_thresholds.setValue(self.session.area_ths)
        self.n_animals.setValue(self.session.number_of_animals)
        self.track_wo_id.setChecked(self.session.track_wo_identities)
        self.check_segm.setChecked(self.session.check_segmentation)
        self.session_name.setText(self.session.name)
        self.bkg_widget.set_bkg_stat(self.session.background_subtraction_stat)
        self.bkg_widget.bkg_thread.n_frames_for_background = (
            self.session.number_of_frames_for_background
        )
        self.bkg_widget.checkBox.setChecked(self.session.use_bkg)
        self.videoPlayer.update()

    def new_parameters(self, params: dict):
        "Receives new toml parameters from OpenVideoWidget"
        unrecognized_params = self.session.set_parameters(**params, reset=True)
        if unrecognized_params:
            QMessageBox.warning(
                self,
                "Error loading parameters",
                f"Unrecognized parameters in toml file:\n\n{unrecognized_params}",
            )
        self.load_parameters()

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self.ROI_Widget.add.setChecked(False)

    def close_and_track_video(self):
        """Action when clicked "close and track video".
        It gathers widgets parameters, writes them in self.user_params and exits"""
        parameters = self.out_parameters()
        if self.unacceptable_parameters(parameters):
            return

        logging.info(pprint_dict(parameters, "GUI params"), extra={"markup": True})
        self.session.set_parameters(**parameters)
        bkg = self.bkg_widget.getBkg()
        if bkg is not None:
            tmp_bkg_path = Path(self.session.video_paths[0]).with_suffix(
                ".tmp_background.png"
            )
            cv2.imencode(".png", bkg)[1].tofile(str(tmp_bkg_path))
            self.session.background_from_segmentation_gui = tmp_bkg_path
        self.run_idtrackerai_after_closing = True
        self.close()

    def getSessionName(self) -> str:
        return (
            self.session_name.text() or self.session_name.placeholderText() or "no_name"
        ).strip()

    def out_parameters(self) -> dict:
        """Generates dict of all widgets content

        Returns
        -------
        dict
            Parameter dict containing all widgets content
        """
        out: dict[str, Any] = {
            "video_paths": self.open_widget.getVideoPaths(),
            "intensity_ths": self.intensity_thresholds.value(),
            "area_ths": self.area_thresholds.value(),
            "tracking_intervals": self.tracking_interval.value(),
            "number_of_animals": self.n_animals.value(),
            "use_bkg": self.bkg_widget.checkBox.isChecked(),
            "check_segmentation": self.check_segm.isChecked(),
            "track_wo_identities": self.track_wo_id.isChecked(),
            "roi_list": self.ROI_Widget.getValue(),
        }

        if self.session_name.text():  # put the name at the first position
            out = {"name": self.session_name.text()} | out

        if self.bkg_widget.checkBox.isChecked():
            out["background_subtraction_stat"] = self.bkg_widget.get_bkg_stat()

        if (
            self.ROI_Widget.exclusive_rois.isVisible()
            and self.ROI_Widget.exclusive_rois.isEnabled()
        ):
            out["exclusive_rois"] = self.ROI_Widget.exclusive_rois.isChecked()
        return out

    def unacceptable_parameters(self, parameters: dict) -> bool:
        if (
            parameters["number_of_animals"] == 0
            and not parameters["track_wo_identities"]
        ):
            QMessageBox.warning(
                self,
                "Missing parameters",
                'Please, define the number of animals in the video or check "Track'
                ' without identities".\n\nEven if tracking without identities,'
                " adding the number of animals is recommended to improve the"
                " individual/crossing blob detection.",
            )
            return True
        if parameters["roi_list"] is not None and len(parameters["roi_list"]) == 0:
            QMessageBox.warning(
                self,
                "Missing parameters",
                "Please, add a region of interest or uncheck the Regions of"
                " interest parameter.",
            )
            return True
        return False

    def save_parameters_func(self):
        parameters = self.out_parameters()

        if self.unacceptable_parameters(parameters):
            return

        file_dialog = QFileDialog()
        settings_key = "save_parameters_filedialog_state"
        state = self.settings.value(settings_key, b"")
        if state:
            file_dialog.restoreState(state)
        fileName, _ = file_dialog.getSaveFileName(
            self,
            "Save parameter file",
            str(Path.cwd() / (self.getSessionName() + ".toml")),
            filter="TOML (*.toml)",
        )
        self.settings.setValue(settings_key, file_dialog.saveState())

        if not fileName:
            return

        with open(fileName, "w", encoding="utf_8") as file:
            for key, value in parameters.items():
                file.write(f"{key} = {toml_format(value)}\n")

    def new_video_paths(
        self,
        video_paths: list[str],
        video_size: tuple[int, int],
        n_frames: int,
        fps: float,
        episodes: list,
    ):
        self.session_name.setPlaceholderText(
            "&".join(Path(path).stem for path in video_paths)
        )
        self.ROI_Widget.set_video_size(video_size)
        self.videoPlayer.setEnabled(False)
        self.tracking_interval.reset(n_frames)
        self.frame_analyzer.drawn_frame = -1
        self.bkg_widget.set_new_video_paths(video_paths, episodes)
        self.ROI_Widget.list.ListChanged.emit()
        self.videoPlayer.update_video_paths(video_paths, n_frames, video_size, fps)

        if not self.enabled:
            for widget in self.list_of_widgets:
                widget.setEnabled(True)
            self.enabled = True
            self.videoPlayer.setEnabled(True)

        self.setWindowTitle(
            f"Segmentation App | {Path(video_paths[0]).name}"
            + (" [...]" if len(video_paths) > 1 else "")
        )
        self.videoPlayer.setEnabled(True)
        self.videoPlayer.update()


def toml_format(value: Any, width: int = 50) -> str:
    """Custom .toml formatter.

    Parameters
    ----------
    value : Any
        The value to format
    width : int, optional
        Maximum line width before wrapping, by default 50

    Returns
    -------
    str
        Formatted value
    """
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float, str)):
        return repr(value)
    if value is None:
        return '""'
    if not value:
        return "[]"
    value = list(value)

    if len(repr(value)) < width:
        return repr(value)

    s = "[\n"
    for item in value:
        s += f"    {repr(item)},\n"
    s += "]"
    return s


class SessionName(QLineEdit):
    "Double click to set the placeholder text as the text"

    def mouseDoubleClickEvent(self, event):
        if not self.text() and self.placeholderText():
            self.setText(self.placeholderText())
        else:
            super().mouseDoubleClickEvent(event)



import logging
from collections.abc import Callable
from pathlib import Path

import cv2
import numpy as np

from idtrackerai import Session
from idtrackerai.GUI_tools import VideoPathHolder
from idtrackerai.utils import create_dir, load_trajectories, track


def draw_collage_frame(
    positions: list[tuple[int, int]],
    miniframes: np.ndarray,
    output_shape: tuple[int, int],
    labels: list[str] | None,
) -> np.ndarray:
    canvas = np.zeros((*output_shape, 3), np.uint8)
    miniframe_size = miniframes.shape[1]
    for cur_id in range(len(miniframes)):
        draw_x, draw_y = positions[cur_id]
        canvas[draw_y : draw_y + miniframe_size, draw_x : draw_x + miniframe_size] = (
            miniframes[cur_id]
        )
        if labels is not None:
            canvas = cv2.putText(
                canvas,
                labels[cur_id],
                (draw_x, draw_y - 5),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 255, 255),
            )
    return canvas


def get_individual_miniframes(
    frame: np.ndarray, ordered_centroid: np.ndarray, size: int
) -> np.ndarray:
    if frame.ndim == 2:
        frame = frame[..., None]
    miniframes = np.zeros((len(ordered_centroid), size, size, 3), np.uint8)
    size2 = size // 2
    for cur_id, (x, y) in enumerate(ordered_centroid):
        if x > 0 and y > 0:
            miniframe = frame[
                max(0, y - size2) : y + size2, max(0, x - size2) : x + size2
            ]

            # with the next slices, the centroid is always in the center of the video,
            # even if the miniframe is cropped
            y_location = (
                slice(None, miniframe.shape[0])
                if y > size2
                else slice(-miniframe.shape[0], None)
            )
            x_location = (
                slice(None, miniframe.shape[1])
                if x > size2
                else slice(-miniframe.shape[1], None)
            )
            miniframes[cur_id, y_location, x_location] = miniframe
    return miniframes


def get_geometries(
    n_animals: int, miniframe_size: int
) -> tuple[list[tuple[int, int]], tuple[int, int]]:
    n_rows = int(np.sqrt(n_animals))
    n_cols = int(np.ceil(n_animals / n_rows))
    extra_lower_pad = 10
    bbox_side_pad = 10
    bbox_top_pad = 30
    full_bbox_width = miniframe_size + 2 * bbox_side_pad
    out_video_width = n_cols * full_bbox_width

    full_bbox_height = miniframe_size + bbox_top_pad
    out_video_height = n_rows * full_bbox_height + extra_lower_pad

    positions = [
        (
            full_bbox_width * (i % n_cols) + bbox_side_pad,
            full_bbox_height * (i // n_cols) + bbox_top_pad,
        )
        for i in range(n_animals)
    ]
    return positions, (out_video_height, out_video_width)


def generate_individual_video(
    session: Session,
    trajectories_path: Path | str | None = None,
    draw_in_gray: bool = False,
    starting_frame: int = 0,
    ending_frame: int | None = None,
    miniframe_size: float | None = None,
    labels: list[str] | None = None,
    callback: Callable | None = None,
    output_dir: Path | str | None = None,
) -> None:
    """
    Generates a set of videos showing the tracked individuals in a session.

    This function creates:
      - A "general" video displaying all individuals in a grid layout, with each animal shown in a square "miniframe".
      - Individual videos for each animal, showing only their own miniframe over time.

    .. seealso::
        Documentation for :ref:`video generator`

    Parameters
    ----------
    session : Session
        The Session instance containing video and tracking data.
    trajectories_path : Path | str | None, optional
        Path to the trajectories file (NumPy .npy or .npz). If None, uses the session's default trajectories folder.
    draw_in_gray : bool, optional
        If True, output videos will be in grayscale. If False, videos will be in color. Default is False.
    starting_frame : int, optional
        Index of the first frame to include in the output videos. Default is 0.
    ending_frame : int | None, optional
        Index of the last frame to include (inclusive). If None, processes until the last frame in the trajectories. Default is None.
    miniframe_size : float | None, optional
        Size (in pixels) of the square miniframe for each individual. If None, uses the session's median body length. Default is None.
    labels : list[str] | None, optional
        Optional list of labels (e.g., animal IDs or names) to display above each miniframe in the general video. Default is None.
    callback : Callable | None, optional
        Optional callback function for progress tracking (e.g., for GUI updates). Default is None.
    output_dir : Path | str | None, optional
        Directory where output videos will be saved. If None, saves to the session folder under "individual_videos". Default is None.

    Output
    ------
    Saves AVI videos to the output directory:
      - "general.avi": grid video of all individuals.
      - "individual_{id}.avi": one video per animal, showing only their own miniframe.

    Notes
    -----
    - The function handles missing or invalid frames by filling with black images.
    - The grid layout and miniframe size are automatically determined based on the number of animals and provided parameters.
    - Uses OpenCV for video writing and image processing.

    """
    if draw_in_gray:
        logging.info("Drawing original video in grayscale")

    trajectories = load_trajectories(trajectories_path or session.trajectories_folder)[
        "trajectories"
    ]

    trajectories = np.nan_to_num(trajectories, nan=-1).astype(int)

    output_dir = (
        Path(output_dir) if output_dir is not None else session.session_folder
    ) / "individual_videos"
    create_dir(output_dir)

    miniframe_size = 2 * (int(miniframe_size or session.median_body_length) // 2)
    logging.info(f"Square individual videos set to {miniframe_size} pixels")
    positions, output_shape = get_geometries(session.n_animals, miniframe_size)

    videoPathHolder = VideoPathHolder(session.video_paths)

    ending_frame = len(trajectories) - 1 if ending_frame is None else ending_frame
    logging.info(f"Drawing from frame {starting_frame} to {ending_frame}")

    general_video_writer = cv2.VideoWriter(
        str(output_dir / "general.avi"),
        cv2.VideoWriter.fourcc(*"XVID"),
        session.frames_per_second,
        output_shape[::-1],
    )

    individual_video_writers = [
        cv2.VideoWriter(
            str(output_dir / f"individual_{id + 1}.avi"),
            cv2.VideoWriter.fourcc(*"XVID"),
            session.frames_per_second,
            (miniframe_size, miniframe_size),
        )
        for id in range(session.n_animals)
    ]

    for frame in track(
        range(starting_frame, ending_frame + 1), "Generating video", callback=callback
    ):
        try:
            img = videoPathHolder.read_frame(frame, not draw_in_gray)
        except RuntimeError as exc:
            logging.error(str(exc))
            img = np.zeros(
                (
                    (session.height, session.width)
                    if draw_in_gray
                    else (session.height, session.width, 3)
                ),
                np.uint8,
            )

        miniframes = get_individual_miniframes(img, trajectories[frame], miniframe_size)

        general_frame = draw_collage_frame(positions, miniframes, output_shape, labels)

        general_video_writer.write(general_frame)

        for id in range(session.n_animals):
            individual_video_writers[id].write(miniframes[id])

    logging.info(f"Videos generated in {output_dir}")



from collections.abc import Generator, Iterable, Sequence
from functools import cached_property
from itertools import chain
from math import atan2, sqrt
from typing import Any, NamedTuple

import cv2
import numpy as np

from .utils import deprecated


class BoundingBoxCoordinates(NamedTuple):
    bottom: int
    left: int
    top: int
    right: int


class Blob:
    """Represents a segmented animal(s) defined with a contour in a specific frame."""

    contour: np.ndarray
    """The coordinates of the contour defining self with shape [n_points, 2]"""

    episode: int

    id_image_index: int
    """Index of the identification image position in the HDF5 file"""

    is_an_individual: bool
    """Flag indicating self represents a single animal."""

    next: tuple["Blob", ...]
    """The :class:`Blob` s from the next frame that overlap with self"""

    previous: tuple["Blob", ...]
    """The :class:`Blob` s from the previous frame that overlap with self"""

    frame_number: int
    """The index of the frame where self belongs to"""

    bbox_img_id: str

    seems_like_individual: bool
    """Unicity condition or not huge area"""

    used_for_training_crossings: bool = False
    """Flag indicating if the blob has been used to train the
    crossing CNN"""

    was_a_crossing: bool = False
    """Flag indicating whether the blob was created after splitting a
    crossing blob during the crossings interpolation process"""

    identity: int | None = None
    """Identity of the blob assigned during the identification process"""

    fragment_identifier: int = -1
    """Indicates the index of the Fragment that contains the blob,
    -1 means no associated Fragment"""

    pixels_are_from_eroded_blob: bool = False

    user_generated_identities: list[int | None] = None  # type: ignore
    """List of identities of the blob some of which might have been give
    by a user during the validation process"""

    user_generated_centroids: list[tuple[float, float] | None] = None  # type: ignore
    """List of centroids generated by the user during the validation
    processes"""

    identity_corrected_solving_jumps: int | None = None
    """Identity of the blob after correcting impossible velocity jumps"""

    identities_corrected_closing_gaps: list | None = None
    """Identity of the blob after crossings interpolation"""

    interpolated_centroids: list | None = None

    added_by_user: bool = False

    forced_crossing: bool = False
    """Indicates if the crossing attribute has been forced by set_individual_with_identity_0_as_crossings()"""

    exclusive_roi: int = -1
    """Exclusive ROI where the blob belongs to"""

    identity_certainty: float = np.nan
    """Confidence of the identity assigned to the blob"""

    def __init__(
        self,
        contour: np.ndarray,
        frame_number: int = -1,
        bbox_img_id: str = "",
        pixels_are_from_eroded_blob: bool = False,
    ) -> None:
        self.set_contour(contour)
        self.frame_number = frame_number
        self.bbox_img_id = bbox_img_id
        if pixels_are_from_eroded_blob:
            self.pixels_are_from_eroded_blob = pixels_are_from_eroded_blob
            # TODO fix this, some eroded blobs are not classified as
            # individuals/crossings and idtrackerai crashes
            self.is_an_individual = True
            self.forced_crossing = True

        self.next = ()
        self.previous = ()

    @property
    def n_next(self) -> int:
        return len(self.next)

    @property
    def n_previous(self) -> int:
        return len(self.previous)

    @cached_property
    def convexHull(self) -> np.ndarray:
        """Convex hull of the contour computed with :func:`cv2.convexHull`"""
        return cv2.convexHull(self.contour)

    @cached_property
    def area(self) -> float:
        """Area of the contour computed with :func:`cv2.contourArea`"""
        return cv2.contourArea(self.contour)

    @cached_property
    def bbox_corners(self) -> BoundingBoxCoordinates:
        """A NamedTuple of the bottom, left, top and right values of the bounding box"""
        return BoundingBoxCoordinates(*self.contour.min(0), *self.contour.max(0))

    @property
    def extension(self) -> float:
        """Extension measured as the length of the diagonal of the bounding box"""
        width, height = np.ptp(self.contour, axis=0)
        return sqrt(width**2 + height**2)

    @cached_property
    def centroid(self) -> tuple[float, float]:
        M = cv2.moments(self.contour)
        try:
            return M["m10"] / M["m00"], M["m01"] / M["m00"]
        except ZeroDivisionError:
            return tuple(self.contour.mean(0))

    @property
    def orientation(self) -> float:
        M = cv2.moments(self.contour)
        try:
            x = M["m10"] / M["m00"]
            y = M["m01"] / M["m00"]
            a = M["m20"] / M["m00"] - x * x
            b = 2 * (M["m11"] / M["m00"] - x * y)
            c = M["m02"] / M["m00"] - y * y
            return 0.5 * atan2(b, a - c)
        except ZeroDivisionError:
            return 0

    def set_contour(self, contour: np.ndarray) -> None:
        if contour.ndim == 3 and contour.shape[1] == 1:
            # OpenCV returns contours as (n_points, 1, 2)
            contour = contour[:, 0]
        self.contour = contour.astype(np.int32, copy=False)

    def __getstate__(self) -> dict[str, Any]:
        out = self.__dict__.copy()
        # clear cached_properties before pickling
        out.pop("convexHull", None)
        out.pop("centroid", None)
        out.pop("area", None)
        out.pop("bbox_corners", None)
        return out

    @property
    def is_a_crossing(self) -> bool:
        """Flag indicating whether the blob represents two or more animals
        together. It is the negative of :attr:`is_an_individual`."""
        return not self.is_an_individual

    @cached_property
    def has_multiple_previous(self) -> bool:
        """Flag indicating if the blob has multiple blobs in its past or future
        overlapping history

        This method is used to check whether the blob is a crossing.

        Returns
        -------
        bool
            If True the blob splits into two or multiple overlapping blobs in
            its "past" or "future" history, depending on the parameter
            "direction"."""
        previous = self
        analyzed_blobs: "list[Blob]" = [previous]
        while previous.n_previous == 1:
            previous = previous.previous[0]
            if "has_multiple_previous" in previous.__dict__:
                # previous has already the answer
                result = previous.has_multiple_previous
                break
            analyzed_blobs.append(previous)
            if previous.n_previous > 1:
                result = True
                break
        else:
            result = False

        for blob in analyzed_blobs:
            blob.has_multiple_previous = result
        return result

    @cached_property
    def has_multiple_next(self) -> bool:
        """Flag indicating if the blob has multiple blobs in its past or future
        overlapping history

        This method is used to check whether the blob is a crossing.

        Returns
        -------
        bool
            If True the blob splits into two or multiple overlapping blobs in
            its "past" or "future" history, depending on the parameter
            "direction".
        """

        next = self
        analyzed_blobs: "list[Blob]" = [next]
        while next.n_next == 1:
            next = next.next[0]
            if "has_multiple_next" in next.__dict__:
                # previous has already the answer
                result = next.has_multiple_next
                break
            analyzed_blobs.append(next)
            if next.n_next > 1:
                result = True
                break
        else:
            result = False

        for blob in analyzed_blobs:
            blob.has_multiple_next = result
        return result

    @cached_property
    def has_a_next_crossing(self) -> bool:
        """Flag indicating if the blob has a crossing in its future overlapping history

        Returns
        -------
        bool
            If True the blob has a crossing in its "future" history
        """
        next = self.next[0]
        analyzed_blobs: "list[Blob]" = [next]
        while next.n_next == 1:
            next = next.next[0]
            if "has_a_next_crossing" in next.__dict__:
                # previous has already the answer
                result = next.has_a_next_crossing
                break
            analyzed_blobs.append(next)
            if next.n_previous > 1 and not next.seems_like_individual:
                result = True
                break
        else:
            result = False

        for blob in analyzed_blobs:
            blob.has_a_next_crossing = result
        return result

    @cached_property
    def has_a_previous_crossing(self) -> bool:
        """Flag indicating if the blob has a crossing in its past overlapping history

        Returns
        -------
        bool
            If True the blob has a crossing in its "past" history
        """

        previous = self.previous[0]
        analyzed_blobs: "list[Blob]" = [previous]
        while previous.n_previous == 1:
            previous = previous.previous[0]
            if "has_a_previous_crossing" in previous.__dict__:
                # previous has already the answer
                result = previous.has_a_previous_crossing
                break
            analyzed_blobs.append(previous)
            if previous.n_next > 1 and not previous.seems_like_individual:
                result = True
                break
        else:
            result = False

        for blob in analyzed_blobs:
            blob.has_a_previous_crossing = result
        return result

    def is_a_sure_individual(self) -> bool:
        """Flag indicating that the blob is a sure individual according to
        some heuristics and it can be used to train the crossing detector CNN.
        """
        return (
            self.seems_like_individual
            and self.n_previous == 1
            and self.n_next == 1
            and self.previous[0].n_next == 1
            and self.next[0].n_previous == 1
            and self.has_a_previous_crossing
            and self.has_a_next_crossing
        )

    def is_a_sure_crossing(self) -> bool:
        """Flag indicating that the blob is a sure crossing according to
        some heuristics and it can be used to train the crossing detector CNN.

        Returns
        -------
        bool
        """
        if self.seems_like_individual:
            return False
        if self.n_previous > 1 or self.n_next > 1:
            return True
        return self.has_multiple_previous and self.has_multiple_next

    def overlaps_with(self, other: "Blob") -> bool:
        """Computes whether the pixels in `self` intersect with the pixels in
        `other`

        Parameters
        ----------
        other : <Blob object>
            An instance of the class Blob

        Returns
        -------
        bool
            True if the lists of pixels of both blobs have non-empty
            intersection
        """

        if (
            self.bbox_corners.top < other.bbox_corners.bottom
            or other.bbox_corners.top < self.bbox_corners.bottom
            or self.bbox_corners.right < other.bbox_corners.left
            or other.bbox_corners.right < self.bbox_corners.left
        ):
            # the bounding boxes do not overlap
            return False

        # Check for every point in `other`'s contour
        points = other.contour.astype(float)
        for point in chain(points[0::3], points[1::3], points[2::3]):
            if self.contour_contains_point(point):
                return True

        # Check for every point in `self`'s contour
        points = self.contour.astype(float)
        return any(
            other.contour_contains_point(point)
            for point in chain(points[0::3], points[1::3], points[2::3])
        )

    def contour_contains_point(self, point: tuple[float, float]) -> bool:
        return cv2.pointPolygonTest(self.contour, point, False) >= 0

    def contains_point(self, point: tuple[float, float]) -> bool:
        if not (
            point[0] >= self.bbox_corners.bottom
            and point[0] <= self.bbox_corners.top
            and point[1] >= self.bbox_corners.left
            and point[1] <= self.bbox_corners.right
        ):
            return False
        return self.contour_contains_point(point)

    def now_points_to(self, other: "Blob") -> None:
        """Given two consecutive blob objects updates their respective
        overlapping histories

        Parameters
        ----------
        other : Blob
            An instance of the class Blob
        """
        self.next = self.next + (other,)
        other.previous = other.previous + (self,)

    def square_distance_to(self, other: "Blob|tuple|list|np.ndarray") -> float:
        """Returns the squared distance from the centroid of self to `other` or the centroid of `other` if it is a Blob.

        Parameters
        ----------
        other : Blob or tuple
            An instance of the class Blob or a tuple (x,y)

        Returns
        -------
        float
            Squared distance between centroids
        """
        if isinstance(other, Blob):
            return ((np.asarray(self.centroid) - np.asarray(other.centroid)) ** 2).sum()

        if isinstance(other, (tuple, list, np.ndarray)):
            return ((np.asarray(self.centroid) - np.asarray(other)) ** 2).sum()

        raise ValueError

    def distance_to(self, other: "Blob|tuple|list|np.ndarray") -> float:
        return sqrt(self.square_distance_to(other))

    def distance_from_countour_to(self, point: Sequence[float]) -> float:
        """Returns the distance between `point` and the closest
        point in the contour of the blob.

        Parameters
        ----------
        point : tuple
            (x,y)

        Returns
        -------
        float
            Smallest distance between `point` and the contour of the blob.
        """
        return abs(cv2.pointPolygonTest(self.contour, point, True))

    @property
    def assigned_identities(self) -> list[None] | list[int] | list[int | None]:
        """Identities assigned to the blob during the tracking process"""
        if self.identities_corrected_closing_gaps is not None:
            return self.identities_corrected_closing_gaps
        if self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        return [self.identity]

    @property
    def assigned_centroids(self) -> list[tuple[float, float]]:
        """Centroids assigned to the blob during the tracking process.

        It considers the default centroid of the blob at segmentation time
        or new centroids added to the blob during the interpolation of the
        crossings.

        Returns
        -------
        list
            List of pairs (x, y) indicating the position of each individual
            in the blob.
        """
        if self.interpolated_centroids:
            return self.interpolated_centroids
        return [self.centroid]

    @property
    def final_centroids(self) -> Generator[tuple[float, float], None, None]:
        """List of the animal/s centroid/s in the blob, considering the
        potential centroids that might have been added by the user during
        the validation.

        By default the centroid will be the center of mass of the blob of
        pixels defined by the blob. It can be different if the user modified
        the default centroid during validation or generated more centroids.

        Returns
        -------
        list
            List of tuples (x, y) indicating the centroids of the blob.
        """
        return (c for c in self.all_final_centroids if c != (-1, -1))

    @property
    def all_final_centroids(self) -> list:
        if self.user_generated_centroids:
            # Note that sometimes len(user_generated_centroids) >
            # len(assigned_centroids)
            final_centroids = []
            for i, centroid in enumerate(self.user_generated_centroids):
                if centroid is not None or i >= len(self.assigned_centroids):
                    final_centroids.append(centroid)
                else:
                    final_centroids.append(self.assigned_centroids[i])
            return final_centroids
        return self.assigned_centroids

    @property
    def final_identities(self) -> Generator[int | None, None, None]:
        return (id for id in self.all_final_identities if id != -1)

    @property
    def all_final_identities(self) -> list:
        """Identities of the blob after the tracking process and after
        potential modifications by the users during the validation procedure.
        """
        if self.user_generated_identities:
            # Note that sometimes len(user_generated_identities)
            # > len(assigned_identities)
            final_identities = []
            # None means the same as assigned, 0 means no id, -1 means no centroid
            for i, user_generated_identity in enumerate(self.user_generated_identities):
                if user_generated_identity is not None or i >= len(
                    self.assigned_identities
                ):
                    final_identities.append(user_generated_identity)
                else:
                    final_identities.append(self.assigned_identities[i])
            return final_identities
        return self.assigned_identities

    @property
    def final_ids_and_centroids(
        self,
    ) -> Iterable[tuple[int | None, tuple[float, float]]]:
        return zip(self.final_identities, self.final_centroids)

    def get_image_for_identification(
        self, img_size: int, bbox_img: np.ndarray, resolution_reduction: float
    ) -> np.ndarray:
        """Generates the image used to train and evaluate the crossing detector CNN and the identification model.

        Parameters
        ----------
        img_size : int
            Size of the identification image. The number of channels is always 1 as images in color are still not considered.
        bbox_img : np.ndarray
            Bounding box image of the blob. This is the image extracted directly from the bounding box, with any size and without the background subtracted.
        resolution_reduction : float
            Resolution reduction to apply when creating the identification image

        Returns
        -------
        ndarray
            Square image with black background used to train the crossings detector and the identification model.
        """

        mask = self.get_bbox_mask()

        mask = cv2.dilate(mask, np.ones((3, 3), np.uint8), iterations=1)

        masked_bbox_image = bbox_img * mask
        bbox_img_height, bbox_img_width = masked_bbox_image.shape
        img_size2 = img_size % 2 + img_size // 2

        center_x = int(
            self.centroid[0] - self.bbox_corners.bottom + 1
        )  # bbox_image_pad

        center_y = int(self.centroid[1] - self.bbox_corners.left + 1)  # bbox_image_pad

        d1 = center_x**2 + center_y**2
        d2 = center_x**2 + (bbox_img_height - center_y) ** 2
        d3 = (bbox_img_width - center_x) ** 2 + center_y**2
        d4 = (bbox_img_width - center_x) ** 2 + (bbox_img_height - center_y) ** 2
        diag = int(sqrt(max((d1, d2, d3, d4))))
        diag = max(diag, img_size2)
        id_img = np.zeros((2 * diag, 2 * diag), np.uint8)
        id_img[
            diag - center_y : diag + bbox_img_height - center_y,
            diag - center_x : diag + bbox_img_width - center_x,
        ] = masked_bbox_image

        M = cv2.getRotationMatrix2D(
            (diag, diag), self.orientation * 180 / np.pi - 45, resolution_reduction
        )

        # old method
        id_img = cv2.warpAffine(
            src=id_img,
            M=M,
            dsize=(diag + img_size2, diag + img_size2),
            borderMode=cv2.BORDER_CONSTANT,
            flags=cv2.INTER_CUBIC,
        )
        return id_img[-img_size:, -img_size:]

        # TODO proposed future method
        # id_img = cv2.warpAffine(
        #     src=id_img,
        #     M=M,
        #     dsize=(diag + img_size, diag + img_size),
        #     borderMode=cv2.BORDER_CONSTANT,
        #     flags=cv2.INTER_CUBIC,
        # )

        # # we build the offset like this to have the minimal ones on the
        # # beginning of the array and be preferably selected by max()
        # origins = [0]
        # for offset in range(img_size2 // 2):
        #     origins += (diag - img_size2 + offset, diag - img_size2 - offset)

        # origin = max(
        #     origins,
        #     key=lambda origin: np.count_nonzero(
        #         id_img[origin : origin + img_size, origin : origin + img_size]
        #     ),
        # )

        # return id_img[origin : origin + img_size, origin : origin + img_size]

    def get_bbox_mask(self) -> np.ndarray:
        """Computes the binary mask with the same size as the bounding box image with ones where the blob is.

        Returns
        -------
        np.ndarray
            Binary np.ndarray image
        """
        base = np.zeros(
            (
                self.bbox_corners.right
                - self.bbox_corners.left
                + 2,  # 2 bbox_image_pads
                self.bbox_corners.top
                - self.bbox_corners.bottom
                + 2,  # 2 bbox_image_pads
            ),
            np.uint8,
        )
        return cv2.fillPoly(
            img=base,
            pts=[self.contour],
            color=[1],
            offset=(
                1 - self.bbox_corners.bottom,  # bbox_image_pad
                1 - self.bbox_corners.left,  # bbox_image_pad
            ),
        )

    def update_centroid(
        self,
        old_centroid: tuple[float, float],
        new_centroid: tuple[float, float],
        identity: int,
    ) -> None:
        """[Validation] Updates the centroid of the blob.

        Parameters
        ----------
        old_centroid : tuple
            Centroid to be updated
        new_centroid : tuple
            Coordinates of the new centroid
        identity : int
            Identity of the centroid to be updated
        """

        self.init_validator_variables()
        index, centroid, dist = self.index_and_centroid_closer_to(
            old_centroid, identity
        )

        self.user_generated_centroids[index] = new_centroid
        self.user_generated_identities[index] = identity

    def init_validator_variables(self) -> None:
        if self.user_generated_centroids is None:
            self.user_generated_centroids: list[tuple[float, float] | None] = [
                None
            ] * len(list(self.final_centroids))
        if self.user_generated_identities is None:
            self.user_generated_identities = [None] * len(list(self.final_identities))

    def index_and_centroid_closer_to(
        self, centroid: tuple, identity: int | None
    ) -> tuple[int, tuple[float, float], float]:
        candidates: list[tuple[int, tuple[float, float], float]] = []
        for indx, (_id, _centroid) in enumerate(
            zip(self.all_final_identities, self.all_final_centroids)
        ):
            if identity not in (None, _id):
                continue
            dist = (_centroid[0] - centroid[0]) ** 2 + (_centroid[1] - centroid[1]) ** 2
            candidates.append((indx, _centroid, dist))
        if not candidates:
            raise ValueError("Centroid not found")

        return min(candidates, key=lambda x: x[2])

    def remove_centroid(self, identity: int, centroid: tuple) -> None:
        """[Validation] Deletes a centroid of the blob.

        Parameters
        ----------
        identity : int
            Identity of the centroid to be deleted
        centroid : tuple
            Centroid to be deleted from the blob
        """

        self.init_validator_variables()
        index, centroid, dist = self.index_and_centroid_closer_to(centroid, identity)

        self.user_generated_centroids[index] = (-1, -1)
        self.user_generated_identities[index] = -1

    def add_centroid(self, centroid: tuple[float, float], identity: int | None) -> None:
        """[Validation] Adds a centroid with a given identity to the blob.

        This method is used in the validation GUI. It is useful to add
        centroids for crossing blobs that are missing some centroids, or to
        individual blobs that should have been classified as crossings and
        are also missing some centroids.

        Parameters
        ----------
        centroid : tuple
            Centroid to be added to the blob
        identity : int
            Identity of the centroid to be added
        """

        self.init_validator_variables()

        self.user_generated_centroids.append(centroid)
        self.user_generated_identities.append(identity)

    def update_identity(
        self,
        old_identity: int | None,
        new_identity: int | None,
        close_to_centroid: tuple[float, float],
    ) -> tuple[float, float]:
        """[Validation] Updates the identity of the blob.

        This method is used during the validation GUI.
        It populates the private attributes `user_generated_identities`
        and `user_generated_centroids`.

        Parameters
        ----------
        new_identity : int
            new value for the identity of the blob
        old_identity : int
            old value of the identity of the blob. It must be specified when the
            blob has multiple identities already assigned.
        centroid : tuple
            centroid which identity must be updated.
        """
        self.init_validator_variables()

        index, centroid, dist = self.index_and_centroid_closer_to(
            close_to_centroid, old_identity
        )

        if new_identity == -1:
            # centroid removals are indicated with new_identity = -1
            # If the identity is removed, the corresponding centroid has to be removed too
            centroid = (-1, -1)

        self.user_generated_identities[index] = new_identity
        self.user_generated_centroids[index] = centroid
        return centroid

    def propagate_identity(
        self,
        old_identity: int | None,
        new_identity: int | None,
        centroid: tuple[float, float],
    ) -> tuple[int, int]:
        """[Validation] Propagates the new identity to next and previous blobs.

        This method called in the validation GUI when the used updates the
        identity of a given blob.

        Returns the frame range where identity changes occurred

        Parameters
        ----------
        old_identity : int
            Previous identity of the blob
        new_identity : int
            New identity of the blob
        centroid : tuple
            [description]
        """

        blobs_stack: list[tuple[Blob, tuple[float, float]]]
        first_frame_modified = self.frame_number
        last_frame_modified = self.frame_number

        blobs_stack = [(next_blob, centroid) for next_blob in self.next]
        while blobs_stack:
            current, previous_centroid = blobs_stack.pop()
            if current.fragment_identifier != self.fragment_identifier:
                continue
            try:
                new_centroid = current.update_identity(
                    old_identity, new_identity, previous_centroid
                )
            except ValueError:  # centroid not found on "current"
                continue
            blobs_stack += [(next_blob, new_centroid) for next_blob in current.next]
            last_frame_modified = current.frame_number

        blobs_stack += [(prev_blob, centroid) for prev_blob in self.previous]
        while blobs_stack:
            current, previous_centroid = blobs_stack.pop()
            if current.fragment_identifier != self.fragment_identifier:
                continue

            try:
                new_centroid = current.update_identity(
                    old_identity, new_identity, previous_centroid
                )
            except ValueError:  # centroid not found on "current"
                continue
            blobs_stack += [(prev_blob, new_centroid) for prev_blob in current.previous]
            first_frame_modified = current.frame_number

        return first_frame_modified, last_frame_modified

    @property
    def summary(self) -> Sequence[str]:
        return (
            (("Individual" if self.is_an_individual else "Crossing") + " Blob")
            + (" (forced)" if self.forced_crossing else ""),
            f"{len(self.contour)} vertices in contour of {self.area:.0f} px area",
            f"In fragment {self.fragment_identifier}",
            f"Linked to {self.n_previous} previous blobs",
            f"Linked to {self.n_next} next blobs",
            ("Used" if self.used_for_training_crossings else "Not used")
            + " for training crossings",
            f"Seems like individual: {self.seems_like_individual}",
            f"Pixels are from eroded blob: {self.pixels_are_from_eroded_blob}",
            f"Predicted identity: {self.identity}",
            f"Corrected solving jumps {self.identity_corrected_solving_jumps}",
            f"Corrected solving gaps: {self.identities_corrected_closing_gaps}",
            f"assigned identities: {self.assigned_identities}",
            f"assigned centroids: {repr_of_list_of_points(self.assigned_centroids)}",
            f"user identities: {self.user_generated_identities}",
            f"user centroids: {repr_of_list_of_points(self.user_generated_centroids)}",
            f"final identities: {list(self.final_identities)}",
            f"final centroids: {repr_of_list_of_points(self.final_centroids)}",
            f"Predicted identity certainty: {self.identity_certainty:.2%}",
        )

    # Deprecated properties for backward compatibility

    @property
    @deprecated(version="6.0.0", reason="Use :meth:`bbox_corners` instead")
    def bbox_in_frame_coordinates(self):
        x0, y0, x1, y1 = self.bbox_corners
        return (x0, y0), (x1, y1)

    @property
    @deprecated(version="6.0.0", reason="Use :meth:`extension` instead")
    def estimated_body_length(self):
        return self.extension

    @deprecated(
        version="6.0.0", reason="Use :meth:`contains_point` instead", action="error"
    )
    def bbox_contains_point(self): ...

    @property
    @deprecated(
        version="6.0.0",
        reason="Use :meth:`all_final_identities` or :meth:`all_final_centroids` instead",
        action="error",
    )
    def all_final_ids_and_centroids(self): ...

    @property
    @deprecated(
        version="6.0.0",
        reason="Check if :meth:`user_generated_identities` is not None instead",
    )
    def has_been_modified(self):
        return self.user_generated_identities is not None

    @property
    @deprecated(version="6.0.0", reason="Use :meth:`summary` instead")
    def properties(self):
        return self.summary


def repr_of_list_of_points(list_of_points) -> str:
    if list_of_points is None:
        return "None"
    list_of_str = [
        f"({point[0]:.1f}, {point[1]:.1f})" if point is not None else "None"
        for point in list_of_points
    ]
    return f"[{', '.join(list_of_str)}]"



import logging
from contextlib import suppress

import cv2
import numpy as np

from idtrackerai import (
    Blob,
    IdtrackeraiError,
    ListOfBlobs,
    ListOfFragments,
    ListOfGlobalFragments,
    Session,
)
from idtrackerai.utils import track


def fragmentation_API(
    session: Session, list_of_blobs: ListOfBlobs
) -> tuple[ListOfFragments, ListOfGlobalFragments]:

    if session.exclusive_rois:
        set_blobs_ROI(list_of_blobs, session.ROI_mask)

    compute_fragment_identifier(list_of_blobs.blobs_in_video)

    list_of_fragments = ListOfFragments.from_fragmented_blobs(
        list_of_blobs.all_blobs, session.n_animals, session.id_images_file_paths
    )
    logging.info(
        f"{list_of_fragments.number_of_fragments} Fragments in total, "
        f"{list_of_fragments.number_of_individual_fragments} individuals and "
        f"{list_of_fragments.number_of_crossing_fragments} crossings"
    )

    with suppress(AttributeError):
        # clean up exclusive ROI attribute, it is on Fragments now
        for blob in list_of_blobs.all_blobs:
            del blob.exclusive_roi

    list_of_global_fragments = ListOfGlobalFragments.from_fragments(
        list_of_blobs.blobs_in_video, list_of_fragments.fragments, session.n_animals
    )
    list_of_fragments.manage_accumulable_non_accumulable_fragments(
        list_of_global_fragments.global_fragments,
        list_of_global_fragments.non_accumulable_global_fragments,
    )

    return list_of_fragments, list_of_global_fragments


def compute_fragment_identifier(blobs_in_video: list[list[Blob]]):
    """Associates a unique fragment identifier to individual blobs
    connected with its next and previous blobs.

    Blobs must be connected and classified as individuals or crossings.
    """
    frame_id = 0
    for blobs_in_frame in track(blobs_in_video, "Fragmenting blobs"):
        for blob in blobs_in_frame:
            if blob.fragment_identifier != -1:
                continue

            blob.fragment_identifier = frame_id
            while (
                len(blob.next) == 1
                and len(blob.next[0].previous) == 1
                and blob.next[0].is_an_individual == blob.is_an_individual
                and blob.next[0].exclusive_roi == blob.exclusive_roi
            ):
                blob = blob.next[0]
                blob.fragment_identifier = frame_id

            frame_id += 1


def set_blobs_ROI(list_of_blobs: ListOfBlobs, mask: np.ndarray | None):
    if mask is None:
        raise IdtrackeraiError(
            "Cannot set exclusive ROIs if there's is not a defined ROI"
        )

    contours = find_exclusive_contours(mask)

    for blob in track(list_of_blobs.all_blobs, "Finding blob's exclusive ROI"):
        blob.exclusive_roi = find_parent_ROI(blob.centroid, contours)


def find_exclusive_contours(
    img: np.ndarray | None,
) -> list[tuple[np.ndarray, list[np.ndarray]]]:
    if img is None:
        return []
    all_cnts, hierarchy = cv2.findContours(img, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    if hierarchy is None:
        return []
    hierarchy = hierarchy[0]
    all_cnts = list(map(np.squeeze, all_cnts))
    n_cnts = len(all_cnts)

    contours: list[tuple[np.ndarray, list[np.ndarray]]] = []
    for index, cnt in enumerate(all_cnts):
        if cnt.size < 4:  # one point contours
            continue
        if hierarchy[index][3] == -1:  # this is a top-level contour
            # check all contours that have the current index as their parent
            holes = [
                all_cnts[i]
                for i in range(n_cnts)
                if hierarchy[i][3] == index and all_cnts[i].size >= 4
            ]
            contours.append((cnt, holes))
    return contours


def find_parent_ROI(
    point: tuple[float, float], contours: list[tuple[np.ndarray, list[np.ndarray]]]
):
    for cont_index, (external, holes) in enumerate(contours):
        if cv2.pointPolygonTest(external, point, False) > 0 and all(
            cv2.pointPolygonTest(hole, point, False) < 0 for hole in holes
        ):
            return cont_index
    raise IdtrackeraiError("Not inside any ROI")



import logging
import multiprocessing
from importlib import metadata

import torch
from torch.backends import mps

from idtrackerai import IdtrackeraiError, conf


def _get_device(user_device: str) -> torch.device:
    """Returns the current available device for PyTorch"""
    try:
        # when mocking torch, metadata.version raises PackageNotFoundError
        logging.debug("Using PyTorch %s", metadata.version("torch"))
    except metadata.PackageNotFoundError:
        logging.debug("Using PyTorch (unknown version)")

    if user_device:
        try:
            device = torch.device(user_device)
        except RuntimeError as exc:
            raise IdtrackeraiError(
                f"Torch device name {user_device!r} not recognized."
            ) from exc
        else:
            logging.info('Using user stated device "%s": %s', user_device, repr(device))
            return device
    if torch.cuda.is_available():
        device = torch.device("cuda")
        logging.info('Using Cuda backend with "%s"', torch.cuda.get_device_name(device))
        return device
    if mps.is_available():
        logging.info("Using MacOS Metal backend")
        return torch.device("mps")

    # Only the main process raises the Warning, spawned processes don't need to insist
    logging.log(
        (
            logging.WARNING
            if multiprocessing.current_process().name == "MainProcess"
            else logging.INFO
        ),
        "[bold red]No graphic device was found available[/], running neural"
        " networks on CPU. This may slow down the training steps.",
        extra={"markup": True},
    )
    return torch.device("cpu")


DEVICE = _get_device(conf.DEVICE)



import random

from qtpy.QtCore import Qt, Signal  # pyright: ignore[reportPrivateImportUsage]
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QApplication,
    QColorDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QMenu,
    QPushButton,
    QScrollArea,
    QWidget,
)

from .other_utils import get_icon


class IdLabels(QScrollArea):
    needToDraw = Signal()
    labels: list[str]

    def __init__(self):
        super().__init__()
        self.grid_layout = QGridLayout()
        self.grid_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setWidgetResizable(True)
        wid = QWidget()
        wid.setLayout(self.grid_layout)
        self.setWidget(wid)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._default_colors = None  # Store default colors

    def change_color(self, idx, btn: QPushButton):
        color = QColorDialog.getColor(
            btn.palette().button().color(),
            self,
            options=QColorDialog.ColorDialogOption.DontUseNativeDialog,
        )
        if not color.isValid():
            return
        set_button_color(btn, color)
        self.colors[idx] = color
        self.transparent_colors[idx] = QColor(
            color.red(), color.green(), color.blue(), 77
        )

    def load(self, labels: list[str], colors: list[str] | list[QColor] | None = None):
        # Remove all widgets from the grid layout
        while self.grid_layout.count():
            item = self.grid_layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()

        self._default_colors = [
            QColor.fromHsv(int(360 * (h / len(labels))), 255, 255)
            for h in range(len(labels))
        ]

        if colors is None:
            colors = self._default_colors

        self.reshuffle_btn = QPushButton()
        self.reshuffle_btn.setIcon(get_icon("refresh"))
        self.reshuffle_btn.setText("Shuffle")
        self.reshuffle_btn.clicked.connect(self.reshuffle_colors)

        self.reset_btn = QPushButton()
        self.reset_btn.setIcon(get_icon("undo"))
        self.reset_btn.setText("Reset")
        self.reset_btn.clicked.connect(self.reset_colors)

        color_btn = QPushButton()
        edit = QLabel("null")
        edit.setEnabled(False)
        edit.setAlignment(Qt.AlignmentFlag.AlignCenter)
        set_button_color(color_btn, QColor(255, 255, 255))
        color_btn.clicked.connect(
            lambda checked, btn=color_btn: self.change_color(0, btn)
        )

        self.grid_layout.addWidget(edit, 1, 1)
        self.grid_layout.addWidget(color_btn, 1, 2)

        self.labels = [""]
        self.colors = [QColor(255, 255, 255)]

        for indx, (label, color) in enumerate(zip(labels, colors), 1):
            if isinstance(color, str):
                color = QColor(color)
            edit = QLineEdit()
            edit.setText(label)
            edit.setPlaceholderText(str(indx))
            edit.setObjectName(str(indx))
            color_btn = QPushButton()
            set_button_color(color_btn, color)
            color_btn.clicked.connect(
                lambda checked, idx=indx, btn=color_btn: self.change_color(idx, btn)
            )
            color_btn.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            color_btn.customContextMenuRequested.connect(
                lambda pos, btn=color_btn, idx=indx: self.show_color_menu(btn, idx, pos)
            )
            edit.textChanged.connect(self.new_label)
            edit.editingFinished.connect(self.validate_label)
            self.grid_layout.addWidget(QLabel(f"{indx}:"), indx + 1, 0)
            self.grid_layout.addWidget(edit, indx + 1, 1)
            self.grid_layout.addWidget(color_btn, indx + 1, 2)

            self.labels.append(label)
            self.colors.append(color)

        self.grid_layout.addWidget(self.reshuffle_btn, len(labels) + 2, 2)
        self.grid_layout.addWidget(self.reset_btn, len(labels) + 3, 2)

        self.transparent_colors = [
            QColor(color.red(), color.green(), color.blue(), 77)
            for color in self.colors
        ]

    def show_color_menu(self, btn: QPushButton, idx: int, pos):
        """Show context menu to copy color code."""
        menu = QMenu(btn)
        hex_code = self.colors[idx].name()
        copy_action = menu.addAction(f"Copy {hex_code}")
        assert copy_action is not None
        copy_action.triggered.connect(
            lambda: QApplication.clipboard().setText(hex_code)
        )
        menu.exec(btn.mapToGlobal(pos))

    def validate_label(self):
        sender = self.sender()
        assert isinstance(sender, QLineEdit)
        text = sender.text()
        if not text:
            sender.setText(sender.placeholderText())
        else:
            sender.setText(text.strip())

    def new_label(self, new_label=""):
        self.labels[int(self.sender().objectName())] = new_label
        self.needToDraw.emit()

    def set_labels_enabled(self, enabled: bool) -> None:
        """Enable or disable label editing."""
        for idx in range(1, len(self.labels)):
            label_edit = self.grid_layout.itemAtPosition(idx + 1, 1)
            if label_edit is not None:
                edit_widget = label_edit.widget()
                if isinstance(edit_widget, QLineEdit):
                    edit_widget.setEnabled(enabled)

    def labels_are_enabled(self) -> bool:
        """Check if label editing is enabled."""
        for idx in range(1, len(self.labels)):
            label_edit = self.grid_layout.itemAtPosition(idx + 1, 1)
            if label_edit is not None:
                edit_widget = label_edit.widget()
                if isinstance(edit_widget, QLineEdit) and not edit_widget.isEnabled():
                    return False
        return True

    def colors_are_enabled(self) -> bool:
        """Check if color buttons are enabled."""
        for idx in range(1, len(self.colors)):
            color_btn = self.grid_layout.itemAtPosition(idx + 1, 2)
            if color_btn is not None:
                btn_widget = color_btn.widget()
                if isinstance(btn_widget, QPushButton) and not btn_widget.isEnabled():
                    return False
        return True

    def set_colors_enabled(self, enabled: bool) -> None:
        """Enable or disable color buttons."""
        for idx in range(len(self.colors)):
            color_btn = self.grid_layout.itemAtPosition(idx + 1, 2)
            if color_btn is not None:
                btn_widget = color_btn.widget()
                if isinstance(btn_widget, QPushButton):
                    btn_widget.setEnabled(enabled)
        self.reshuffle_btn.setEnabled(enabled)
        self.reset_btn.setEnabled(enabled)

    def get_labels(self) -> list[str]:
        return self.labels

    def get_colors(self) -> tuple[list[QColor], list[QColor]]:
        return self.colors, self.transparent_colors

    def reshuffle_colors(self) -> None:
        """Reshuffle the existing colors among the labels and update the UI."""
        if not hasattr(self, "colors") or not self.colors:
            return

        # Do not reshuffle the first color (index 0), which is for 'null'
        label_colors = self.colors[1:]
        random.shuffle(label_colors)
        self.colors[1:] = label_colors
        self.transparent_colors = [
            QColor(c.red(), c.green(), c.blue(), 77) for c in self.colors
        ]

        # Update color buttons in the grid layout
        for idx, color in enumerate(self.colors):
            row = idx + 1  # row 0 is the shuffle button
            color_btn = self.grid_layout.itemAtPosition(row, 2)
            if color_btn is not None:
                btn_widget = color_btn.widget()
                if isinstance(btn_widget, QPushButton):
                    set_button_color(btn_widget, color)

        self.needToDraw.emit()

    def reset_colors(self):
        """Reset colors to their default values and update the UI."""
        if not self._default_colors:
            return
        self.colors = [QColor(255, 255, 255)] + self._default_colors
        self.transparent_colors = [
            QColor(c.red(), c.green(), c.blue(), 77) for c in self.colors
        ]

        for idx, color in enumerate(self.colors):
            row = idx + 1  # row 0 is the buttons
            color_btn = self.grid_layout.itemAtPosition(row, 2)
            if color_btn is not None:
                btn_widget = color_btn.widget()
                if isinstance(btn_widget, QPushButton):
                    set_button_color(btn_widget, color)
        self.needToDraw.emit()


def set_button_color(button: QPushButton, color: QColor):
    """Set the background color of a QPushButton."""
    unsaturated_color = QColor(color)
    unsaturated_color.setHsv(
        unsaturated_color.hue(),
        int(unsaturated_color.saturation() * 0.5),
        unsaturated_color.value(),
        100,
    )
    button.setStyleSheet(
        f"""
        QPushButton {{
            background-color: {color.name()};
            border: black;
            border-width: 1px;
            border-style: solid;
            border-radius: 5px;
            padding: 5px;
            color: black;
        }}
        QPushButton:hover {{
            background-color: {color.lighter(170).name()};
        }}
        QPushButton:disabled {{
            background-color: rgba({unsaturated_color.red()}, {unsaturated_color.green()}, {unsaturated_color.blue()}, {unsaturated_color.alpha()});
            color: gray;
            border: gray;
        }}
        """
    )
    button.setText(color.name())
